"""
Py-ART: The Python ARM Radar Toolkit
=====================================

"""

# print information on citing Py-ART, this message can be suppressed by
# setting the PYART_QUIET environment variable
_citation_text = """
## You are using the Python ARM Radar Toolkit (Py-ART), an open source
## library for working with weather radar data. Py-ART is partly
## supported by the U.S. Department of Energy as part of the Atmospheric
## Radiation Measurement (ARM) Climate Research Facility, an Office of
## Science user facility.
##
## If you use this software to prepare a publication, please cite:
##
##     JJ Helmus and SM Collis, JORS 2016, doi: 10.5334/jors.119
"""
from os import environ as _environ

if "PYART_QUIET" not in _environ:
    print(_citation_text)

from pkg_resources import DistributionNotFound, get_distribution

# import subpackages
# print out helpful message if build fails or importing from source tree
from . import __check_build  # noqa
from . import aux_io  # noqa
from . import bridge  # noqa
from . import config  # noqa
from . import core  # noqa
from . import correct  # noqa
from . import filters  # noqa
from . import graph  # noqa
from . import io  # noqa
from . import map  # noqa
from . import retrieve  # noqa
from . import testing  # noqa
from . import util  # noqa
from ._debug_info import _debug_info  # noqa

# root level functions
from .config import load_config  # noqa

# Get the version
try:
    __version__ = get_distribution("arm_pyart").version
except DistributionNotFound:
    # package is not installed
    __version__ = "0.0.0"



"""
======================================================
Create a multiple panel RHI plot from a CF/Radial file
======================================================

An example which creates a multiple panel RHI plot of a CF/Radial file using
a RadarDisplay object.

"""
print(__doc__)

# Author: Jonathan J. Helmus (jhelmus@anl.gov)
# License: BSD 3 clause

import matplotlib.pyplot as plt
import netCDF4

import pyart
from pyart.testing import get_test_data

filename = get_test_data("sgpxsaprrhicmacI5.c0.20110524.015604_NC4.nc")

# create the plot using RadarDisplay
radar = pyart.io.read_cfradial(filename)
radar.metadata["instrument_name"] = "XSARP"
display = pyart.graph.RadarDisplay(radar)

fig = plt.figure(figsize=[12, 17])
fig.subplots_adjust(hspace=0.4)
xlabel = "Distance from radar (km)"
ylabel = "Height agl (km)"
colorbar_label = "Hz. Eq. Refl. Fac. (dBZ)"
nplots = radar.nsweeps

for snum in radar.sweep_number["data"]:

    fixed_angle = radar.fixed_angle["data"][snum]
    title = "HSRHI Az=%.3f" % (fixed_angle)
    ax = fig.add_subplot(nplots, 1, snum + 1)
    display.plot(
        "reflectivity_horizontal",
        snum,
        vmin=-20,
        vmax=20,
        mask_outside=False,
        title=title,
        axislabels=(xlabel, ylabel),
        colorbar_label=colorbar_label,
        ax=ax,
    )
    display.set_limits(ylim=[0, 15], ax=ax)

time_start = netCDF4.num2date(
    radar.time["data"][0],
    radar.time["units"],
    only_use_cftime_datetimes=False,
    only_use_python_datetimes=True,
)
figure_title = "Time: " + time_start.isoformat() + "Z"
fig.text(0.35, 0.92, figure_title)

plt.show()



"""
Function for extracting the radar column above a target
given position in latitude, longitude

"""

import numpy as np
import pandas as pd
import xarray as xr

from ..core.transforms import antenna_vectors_to_cartesian


def sphere_distance(radar_latitude, target_latitude, radar_longitude, target_longitude):
    """
    Calculated of the great circle distance between radar and target

    Assumptions
    -----------
    Radius of the Earth = 6371 km / 6371000 meters
    Distance is calculated for a smooth sphere
    Radar and Target are at the same altitude (need to check)

    Parameter
    ---------
    radar_latitude : float, [degrees]
        latitude of the radar in degrees
    target_latitude : float, [degrees]
        latitude of the target in degrees
    radar_longitude : float, [degrees]
        longitude of the radar in degrees
    target_longitude : float, [degrees]
        longitude of the target in degress

    Returns
    -------
    distance : float, [meters]
        Great-Circle Distance between radar and target in meters
    """
    # check if latitude, longitudes are valid
    check_latitude(radar_latitude)
    check_latitude(target_latitude)
    check_longitude(radar_longitude)
    check_longitude(target_longitude)

    # convert latitudeitude/longitudegitudes to radians
    radar_latitude = radar_latitude * (np.pi / 180.0)
    target_latitude = target_latitude * (np.pi / 180.0)
    radar_longitude = radar_longitude * (np.pi / 180.0)
    target_longitude = target_longitude * (np.pi / 180.0)

    # difference in latitudeitude
    d_latitude = target_latitude - radar_latitude
    # difference in longitudegitude
    d_longitude = target_longitude - radar_longitude

    # Haversine formula
    numerator = (np.sin(d_latitude / 2.0) ** 2.0) + np.cos(radar_latitude) * np.cos(
        target_latitude
    ) * (np.sin(d_longitude / 2.0) ** 2.0)
    distance = 2 * 6371000 * np.arcsin(np.sqrt(numerator))

    # return the output
    return distance


def for_azimuth(radar_latitude, target_latitude, radar_longitude, target_longitude):
    """
    Calculation of inital bearing alongitudeg a great-circle arc
    Known as Forward Azimuth Angle.

    Assumptions
    -----------
    Radius of the Earth = 6371 km / 6371000 meters
    Distance is calculatitudeed for a smooth sphere
    Radar and Target are at the same altitude (need to check)

    Parameters
    ----------
    radar_latitude : float, [degrees]
        latitude of the radar in degrees
    target_latitude : float, [degrees]
        latitude of the target in degrees
    radar_longitude : float, [degrees]
        longitude of the radar in degrees
    target_longitude : float, [degrees]
        longitude of the target in degress

    Returns
    -------
    azimuth : float, [degrees]
        azimuth angle from the radar where
        target is located within the scan.
        output is in degrees.
    """
    # check the input latitude/longitude
    check_latitude(radar_latitude)
    check_latitude(target_latitude)
    check_longitude(radar_longitude)
    check_longitude(target_longitude)

    # convert latitudeitude/longitudegitudes to radians
    radar_latitude = radar_latitude * (np.pi / 180.0)
    target_latitude = target_latitude * (np.pi / 180.0)
    radar_longitude = radar_longitude * (np.pi / 180.0)
    target_longitude = target_longitude * (np.pi / 180.0)

    # Differnce in longitudegitudes
    d_longitude = target_longitude - radar_longitude

    # Determine x,y coordinates for arc tangent function
    corr_y = np.sin(d_longitude) * np.cos(target_latitude)
    corr_x = (np.cos(radar_latitude) * np.sin(target_latitude)) - (
        np.sin(radar_latitude) * (np.cos(target_latitude) * np.cos(d_longitude))
    )

    # Determine forward azimuth angle
    azimuth = np.arctan2(corr_y, corr_x) * (180.0 / np.pi)

    # Return the output as a function of 0-360 degrees
    if azimuth < 0:
        azimuth += 360.0

    return azimuth


def get_field_location(radar, latitude, longitude):

    """
    Given the location (in latitude, longitude) of a target, extract the
    radar column above that point for further analysis.

    Parameters
    ----------
    radar : pyart.core.Radar Object
        Py-ART Radar Object from which distance to the target, along
        with gates above the target, will be calculated.
    latitude : float, [degrees]
        Latitude, in degrees North, of the target.
    longitude : float, [degrees]
        Longitude, in degrees East, of the target.

    Function Calls
    --------------
    sphere_distance
    for_azimuth
    get_column_rays

    Returns
    -------
    column : xarray DataSet
        Xarary Dataset containing the radar column above the target for
        the various fields within the radar object.
    """
    # Make sure latitude, longitudes are valid
    check_latitude(latitude)
    check_longitude(longitude)

    # initiate a dictionary to hold the gates above the radar.
    zgate = []

    # initiate a diciontary to hold the moment data.
    moment = {key: [] for key in radar.fields.keys()}

    # call the sphere_distance function
    dis = sphere_distance(
        radar.latitude["data"][0], latitude, radar.longitude["data"][0], longitude
    )

    # call the for_azimuth function
    azim = for_azimuth(
        radar.latitude["data"][0], latitude, radar.longitude["data"][0], longitude
    )

    # call the get_column_ray function
    ray = get_column_rays(radar, azim)

    # Determine the center of each gate for the subsetted rays.
    (rhi_x, rhi_y, rhi_z) = antenna_vectors_to_cartesian(
        radar.range["data"],
        radar.azimuth["data"][ray],
        radar.elevation["data"][ray],
        edges=False,
    )
    # Calculate distance from the x,y coordinates to target
    rhidis = np.sqrt((rhi_x**2) + (rhi_y**2)) * np.sign(rhi_z)
    for i in range(len(ray)):
        tar_gate = np.argmin(abs(rhidis[i, 1:] - (dis)))
        for key in moment:
            if radar.fields[key]["data"][ray[i], tar_gate] is np.ma.masked:
                moment[key].append(np.nan)
            else:
                moment[key].append(radar.fields[key]["data"][ray[i], tar_gate])
        # Add radar elevation to height gates
        # to define height as center of each gate above sea level
        zgate.append(rhi_z[i, tar_gate] + radar.altitude["data"][0])

    # Determine the time at the center of each ray within the column
    # Define the start of the radar volume as a numpy datetime object for xr
    # We take advantage of the "seconds since " portion of the units string
    base_time = pd.to_datetime(radar.time["units"][14:]).to_numpy()
    # Convert Py-ART radar object time (time since volume start) to time delta
    # Add to base time to have sequential time within the xr Dataset
    # for easier future merging/work
    combined_time = []
    for i in range(len(ray)):
        delta = pd.to_timedelta(radar.time["data"][ray[i]], unit="s")
        total_time = base_time + delta
        combined_time.append(total_time.to_numpy())

    # Create a blank list to hold the xarray DataArrays
    ds_container = []
    da_meta = [
        "units",
        "standard_name",
        "long_name",
        "valid_max",
        "valid_min",
        "coordinates",
    ]
    # Convert the moment dictionary to xarray DataArray.
    # Apply radar object meta data to DataArray attribute
    for key in moment:
        if key != "height":
            da = xr.DataArray(
                moment[key], coords=dict(height=zgate), name=key, dims=["height"]
            )
            for tag in da_meta:
                if tag in radar.fields[key]:
                    da.attrs[tag] = radar.fields[key][tag]
            # Append to ds container
            ds_container.append(da.to_dataset(name=key))

    # Add additional DataArrays 'base_time' and 'time_offset'
    # if not present within the radar object.
    da_base = xr.DataArray(base_time, name="base_time")
    da_offset = xr.DataArray(
        combined_time, coords=dict(height=zgate), name="time_offset", dims=["height"]
    )
    ds_container.append(da_base.to_dataset(name="base_time"))
    ds_container.append(da_offset.to_dataset(name="time_offset"))

    # Create a xarray DataSet from the DataArrays
    column = xr.merge(ds_container)

    # Assign Attributes for the Height and Times
    height_des = (
        "Height Above Sea Level [in meters] for the Center of Each"
        + " Radar Gate Above the Target Location"
    )
    column.height.attrs.update(
        long_name="Height of Radar Beam",
        units="m",
        standard_name="height",
        description=height_des,
    )

    column.base_time.attrs.update(long_name="UTC Reference Time", units="seconds")

    time_long = "Time in Seconds Since Volume Start"
    time_des = (
        "Time in Seconds Since Volume Start that Cooresponds"
        + " to the Center of Each Height Gate"
        + " Above the Target Location"
    )
    column.time_offset.attrs.update(
        long_name=time_long, units="seconds", description=time_des
    )

    # Assign Global Attributes to the DataSet
    column.attrs["distance_from_radar"] = str(np.around(dis / 1000.0, 3)) + " km"
    column.attrs["azimuth"] = str(np.around(azim, 3)) + " degrees"
    column.attrs["latitude_of_location"] = str(latitude) + " degrees"
    column.attrs["longitude_of_location"] = str(longitude) + " degrees"
    return column


def get_column_rays(radar, azimuth):

    """
    Given the location (in latitude,longitude) of a target, return the rays
    that correspond to radar column above the target.

    Parameters
    ----------
    radar : Radar Object
        Py-ART Radar Object from which distance to the target, along
        with gates above the target, will be calculated.
    azimuth : float,int
        forward azimuth angle from radar to target in degrees.

    Returns
    -------
    nrays : List
        radar ray indices that correspond to the column above a
        target location.
    """
    if isinstance(azimuth, int) or isinstance(azimuth, float) is True:
        if (azimuth <= 0) or (azimuth >= 360):
            raise ValueError("azimuth not valid (not between 0-360 degrees)")
    else:
        raise TypeError(
            "radar azimuth type not valid."
            " Please convert input to be an int or float."
        )
    # define a list to hold the valid rays
    rays = []
    # check to see which radar scan
    if radar.scan_type == "rhi":
        for i in range(radar.sweep_number["data"].shape[0]):
            nstart = radar.sweep_start_ray_index["data"][i]
            nstop = radar.sweep_end_ray_index["data"][i]
            counter = 0
            for j in range(nstart, nstop):
                if abs(radar.azimuth["data"][nstart + counter] - azimuth) < 1:
                    rays.append(nstart + counter)
                counter += 1
    else:
        # taken from pyart.graph.RadarDisplay.get_azimuth_rhi_data_x_y_z
        for sweep in radar.iter_slice():
            sweep_azi = radar.azimuth["data"][sweep]
            nray = np.argmin(np.abs(sweep_azi - azimuth))
            rays.append(nray + sweep.start)
    # make sure rays were found
    if len(rays) == 0:
        raise ValueError("No rays were found between azimuth and target")

    return rays


def check_latitude(latitude):

    """
    Function to check if input latitude is valid for type and value.

    Parameters
    ----------
    latitude : int, float
        Latitude of a location that should be between 90S and 90N

    """
    if (
        isinstance(latitude, int)
        or isinstance(latitude, float)
        or isinstance(latitude, np.floating)
    ) is True:
        if (latitude <= -90) or (latitude >= 90):
            raise ValueError(
                "Latitude not between -90 and 90 degrees, need to "
                "convert to values between -90 and 90"
            )
    else:
        raise TypeError(
            "Latitude type not valid, need to convert input to be an int or float"
        )


def check_longitude(longitude):

    """
    Function to check if input latitude is valid for type and value.

    Parameters
    ---------
    longitude : int, float
        Longitude of a location taht should be between 180W and 180E
    """
    if (
        isinstance(longitude, int)
        or isinstance(longitude, float)
        or isinstance(longitude, np.floating)
    ) is True:
        if (longitude <= -180) or (longitude >= 180):
            raise ValueError(
                "Longitude not valid between -180 and 180"
                " degrees, need to convert to values between"
                "  -180 and 180"
            )
    else:
        raise TypeError(
            "Longitude type not valid, need to convert input to be an int or float"
        )



""" Unit Tests for Py-ART's io/mdv.py module. """


import numpy as np
import pytest

import pyart


@pytest.mark.skipif(
    not pyart.bridge.wradlib_bridge._WRADLIB_AVAILABLE,
    reason="Wradlib is not installed.",
)
def test_texture_of_complex_phase():
    test_radar = pyart.testing.make_empty_ppi_radar(100, 360, 5)
    foo_field = {"data": np.zeros([360 * 5, 100])}
    test_radar.add_field("differential_phase", foo_field)
    test_text = pyart.retrieve.texture_of_complex_phase(test_radar)
    assert test_text["data"].mean() == 0.0



"""
A general central radial scanning (or dwelling) instrument class.

"""

import warnings

import numpy as np

try:
    import xarray as xr

    _XARRAY_AVAILABLE = True
except ImportError:
    _XARRAY_AVAILABLE = False

from ..config import get_metadata
from ..exceptions import MissingOptionalDependency
from .transforms import antenna_vectors_to_cartesian, cartesian_to_geographic


class RadarSpectra:
    """
    A class for storing antenna coordinate radar spectra data.
    The structure of the Radar class is based on the CF/Radial Data file
    format. Global attributes and variables (section 4.1 and 4.3) are
    represented as a dictionary in the metadata attribute. Other required and
    optional variables are represented as dictionaries in a attribute with the
    same name as the variable in the CF/Radial standard. When a optional
    attribute not present the attribute has a value of None. The data for a
    given variable is stored in the dictionary under the 'data' key. Moment
    field data is stored as a dictionary of dictionaries in the fields
    attribute. Sub-convention variables are stored as a dictionary of
    dictionaries under the meta_group attribute.

    Refer to the attribute section for information on the parameters.

    Attributes
    ----------
    time : dict
        Time at the center of each ray.
    range : dict
        Range to the center of each gate (bin).
    fields : dict of dicts
        Spectra fields.
    metadata : dict
        Metadata describing the instrument and data.
    scan_type : str
        Type of scan, one of 'ppi', 'rhi', 'sector' or 'other'. If the scan
        volume contains multiple sweep modes this should be 'other'.
    latitude : dict
        Latitude of the instrument.
    longitude : dict
        Longitude of the instrument.
    altitude : dict
        Altitude of the instrument, above sea level.
    altitude_agl : dict or None
        Altitude of the instrument above ground level. If not provided this
        attribute is set to None, indicating this parameter not available.
    sweep_number : dict
        The number of the sweep in the volume scan, 0-based.
    sweep_mode : dict
        Sweep mode for each mode in the volume scan.
    fixed_angle : dict
        Target angle for thr sweep. Azimuth angle in RHI modes, elevation
        angle in all other modes.
    sweep_start_ray_index : dict
        Index of the first ray in each sweep relative to the start of the
        volume, 0-based.
    sweep_end_ray_index : dict
        Index of the last ray in each sweep relative to the start of the
        volume, 0-based.
    rays_per_sweep : LazyLoadDict
        Number of rays in each sweep. The data key of this attribute is
        create upon first access from the data in the sweep_start_ray_index and
        sweep_end_ray_index attributes. If the sweep locations needs to be
        modified, do this prior to accessing this attribute or use
        :py:func:`init_rays_per_sweep` to reset the attribute.
    target_scan_rate : dict or None
        Intended scan rate for each sweep. If not provided this attribute is
        set to None, indicating this parameter is not available.
    rays_are_indexed : dict or None
        Indication of whether ray angles are indexed to a regular grid in
        each sweep. If not provided this attribute is set to None, indicating
        ray angle spacing is not determined.
    ray_angle_res : dict or None
        If rays_are_indexed is not None, this provides the angular resolution
        of the grid. If not provided or available this attribute is set to
        None.
    azimuth : dict
        Azimuth of antenna, relative to true North. Azimuth angles are
        recommended to be expressed in the range of [0, 360], but other
        representations are not forbidden.
    elevation : dict
        Elevation of antenna, relative to the horizontal plane. Elevation
        angles are recommended to be expressed in the range of [-180, 180],
        but other representations are not forbidden.
    gate_x, gate_y, gate_z : LazyLoadDict
        Location of each gate in a Cartesian coordinate system assuming a
        standard atmosphere with a 4/3 Earth's radius model. The data keys of
        these attributes are create upon first access from the data in the
        range, azimuth and elevation attributes. If these attributes are
        changed use :py:func:`init_gate_x_y_z` to reset.
    gate_longitude, gate_latitude : LazyLoadDict
        Geographic location of each gate. The projection parameter(s) defined
        in the `projection` attribute are used to perform an inverse map
        projection from the Cartesian gate locations relative to the radar
        location to longitudes and latitudes. If these attributes are changed
        use :py:func:`init_gate_longitude_latitude` to reset the attributes.
    projection : dic or str
        Projection parameters defining the map projection used to transform
        from Cartesian to geographic coordinates. The default dictionary sets
        the 'proj' key to 'pyart_aeqd' indicating that the native Py-ART
        azimuthal equidistant projection is used. This can be modified to
        specify a valid pyproj.Proj projparams dictionary or string.
        The special key '_include_lon_0_lat_0' is removed when interpreting
        this dictionary. If this key is present and set to True, which is
        required when proj='pyart_aeqd', then the radar longitude and
        latitude will be added to the dictionary as 'lon_0' and 'lat_0'.
    gate_altitude : LazyLoadDict
        The altitude of each radar gate as calculated from the altitude of the
        radar and the Cartesian z location of each gate. If this attribute
        is changed use :py:func:`init_gate_altitude` to reset the attribute.
    scan_rate : dict or None
        Actual antenna scan rate. If not provided this attribute is set to
        None, indicating this parameter is not available.
    antenna_transition : dict or None
        Flag indicating if the antenna is in transition, 1 = yes, 0 = no.
        If not provided this attribute is set to None, indicating this
        parameter is not available.
    rotation : dict or None
        The rotation angle of the antenna. The angle about the aircraft
        longitudinal axis for a vertically scanning radar.
    tilt : dict or None
        The tilt angle with respect to the plane orthogonal (Z-axis) to
        aircraft longitudinal axis.
    roll : dict or None
        The roll angle of platform, for aircraft right wing down is positive.
    drift : dict or None
        Drift angle of antenna, the angle between heading and track.
    heading : dict or None
        Heading (compass) angle, clockwise from north.
    pitch : dict or None
        Pitch angle of antenna, for aircraft nose up is positive.
    georefs_applied : dict or None
        Indicates whether the variables have had georeference calculation
        applied.  Leading to Earth-centric azimuth and elevation angles.
    instrument_parameters : dict of dicts or None
        Instrument parameters, if not provided this attribute is set to None,
        indicating these parameters are not avaiable. This dictionary also
        includes variables in the radar_parameters CF/Radial subconvention.
    radar_calibration : dict of dicts or None
        Instrument calibration parameters. If not provided this attribute is
        set to None, indicating these parameters are not available
    ngates : int
        Number of gates (bins) in a ray.
    nrays : int
        Number of rays in the volume.
    nsweeps : int
        Number of sweep in the volume.

    """

    def __init__(
        self,
        time,
        _range,
        fields,
        metadata,
        scan_type,
        latitude,
        longitude,
        altitude,
        sweep_number,
        sweep_mode,
        fixed_angle,
        sweep_start_ray_index,
        sweep_end_ray_index,
        azimuth,
        elevation,
        npulses_max,
        velocity_bins,
        altitude_agl=None,
        target_scan_rate=None,
        rays_are_indexed=None,
        ray_angle_res=None,
        scan_rate=None,
        antenna_transition=None,
        instrument_parameters=None,
        radar_calibration=None,
        georefs_applied=None,
    ):
        warnings.warn(
            "Radar Spectra object is in early development, "
            "errors may arise, use at your own risk! "
        )
        if not _XARRAY_AVAILABLE:
            raise MissingOptionalDependency(
                "Xarray is required to use RadarSpectra but is " "not installed!"
            )
        self.field_names = ["spectra"]

        self.ds = xr.Dataset(
            data_vars={
                "spectra": (("time", "range", "npulses_max"), fields),
                "velocity_bins": velocity_bins,
                "scan_type": scan_type,
                "latitude": latitude,
                "longitude": longitude,
                "altitude": altitude,
                "sweep_number": sweep_number,
                "sweep_mode": sweep_mode,
                "fixed_angle": fixed_angle,
                "sweep_start_ray_index": sweep_start_ray_index,
                "sweep_end_ray_index": sweep_end_ray_index,
                "azimuth": azimuth,
                "elevation": elevation,
            },
            coords={"time": time, "range": _range, "npulses_max": npulses_max},
            attrs=metadata,
        )

        self.ds["ngates"] = len(_range.values)
        self.ds["nrays"] = len(time.values)
        self.ds["nsweeps"] = len(sweep_number.values)
        self.ds.attrs["projection"] = {
            "proj": "pyart_aeqd",
            "_include_lon_0_lat_0": True,
        }

        # initalize attributes with lazy load dictionaries
        self.init_rays_per_sweep()
        self.init_gate_x_y_z()
        self.init_gate_longitude_latitude()
        self.init_gate_altitude()

    @property
    def fields(self):
        field_dict = {}
        for key in self.field_names:
            if key in self.ds.variables.keys():
                field_dict[key] = self.ds[key]
        return xr.Dataset(field_dict)

    @property
    def time(self):
        return self.ds.time

    @property
    def range(self):
        return self.ds.range

    @property
    def npulses_max(self):
        return self.ds.npulses_max

    @property
    def velocity_bins(self):
        return self.ds.velocity_bins

    @property
    def latitude(self):
        return self.ds.latitude

    @property
    def longitude(self):
        return self.ds.longitude

    @property
    def altitude(self):
        return self.ds.altitude

    @property
    def fixed_angle(self):
        return self.ds.fixed_angle

    @property
    def sweep_mode(self):
        return self.ds.sweep_mode

    @property
    def sweep_number(self):
        return self.ds.sweep_number

    @property
    def scan_type(self):
        return self.ds.scan_type

    @property
    def elevation(self):
        return self.ds.elevation

    @property
    def azimuth(self):
        return self.ds.azimuth

    @property
    def sweep_start_ray_index(self):
        return self.ds.sweep_start_ray_index

    @property
    def sweep_end_ray_index(self):
        return self.ds.sweep_end_ray_index

    @property
    def rays_per_sweep(self):
        return self.ds.rays_per_sweep

    @property
    def gate_x(self):
        return self.ds.gate_x

    @property
    def gate_y(self):
        return self.ds.gate_y

    @property
    def gate_z(self):
        return self.ds.gate_z

    @property
    def gate_latitude(self):
        return self.ds.gate_latitude

    @property
    def gate_longitude(self):
        return self.ds.gate_longitude

    @property
    def gate_altitude(self):
        return self.ds.gate_altitude

    @property
    def ngates(self):
        return self.ds.ngates

    @property
    def nrays(self):
        return self.ds.nrays

    @property
    def nsweeps(self):
        return self.ds.nsweeps

    @property
    def projection(self):
        return self.ds.attrs["projection"]

    def init_rays_per_sweep(self):
        """Initialize or reset the rays_per_sweep attribute."""
        _rays_per_sweep_data_factory(self.ds)

    def init_gate_x_y_z(self):
        """Initialize or reset the gate_{x, y, z} attributes."""
        _gate_data_factory(self.ds)

    def init_gate_longitude_latitude(self):
        """
        Initialize or reset the gate_longitude and gate_latitude attributes.
        """
        _gate_lon_lat_data_factory(self.ds)

    def init_gate_altitude(self):
        """Initialize the gate_altitude attribute."""
        _gate_altitude_data_factory(self.ds)

    def _check_sweep_in_range(self, sweep):
        """Check that a sweep number is in range."""
        if sweep < 0 or sweep >= self.nsweeps:
            raise IndexError("Sweep out of range: ", sweep)

        # get methods

    def get_start(self, sweep):
        """Return the starting ray index for a given sweep."""
        self._check_sweep_in_range(sweep)
        return self.sweep_start_ray_index.values[sweep]

    def get_end(self, sweep):
        """Return the ending ray for a given sweep."""
        self._check_sweep_in_range(sweep)
        return self.sweep_end_ray_index.values[sweep]

    def get_start_end(self, sweep):
        """Return the starting and ending ray for a given sweep."""
        return self.get_start(sweep), self.get_end(sweep)

    def get_slice(self, sweep):
        """Return a slice for selecting rays for a given sweep."""
        start, end = self.get_start_end(sweep)
        return slice(start, end + 1)

    def check_field_exists(self, field_name):
        """
        Check that a field exists in the fields dictionary.
        If the field does not exist raise a KeyError.

        Parameters
        ----------
        field_name : str
            Name of field to check.

        """
        if field_name not in self.fields.keys():
            raise KeyError("Field not available: " + field_name)

    # Iterators
    def iter_start(self):
        """Return an iterator over the sweep start indices."""
        return (s for s in self.sweep_start_ray_index.values)

    def iter_end(self):
        """Return an iterator over the sweep end indices."""
        return (s for s in self.sweep_end_ray_index.values)

    def iter_start_end(self):
        """Return an iterator over the sweep start and end indices."""
        return ((s, e) for s, e in zip(self.iter_start(), self.iter_end()))

    def iter_slice(self):
        """Return an iterator which returns sweep slice objects."""
        return (slice(s, e + 1) for s, e in self.iter_start_end())

    def iter_field(self, field_name):
        """Return an iterator which returns sweep field data."""
        self.check_field_exists(field_name)
        return (self.fields[field_name].values[s] for s in self.iter_slice())

    def iter_azimuth(self):
        """Return an iterator which returns sweep azimuth data."""
        return (self.azimuth.values[s] for s in self.iter_slice())

    def iter_elevation(self):
        """Return an iterator which returns sweep elevation data."""
        return (self.elevation.values[s] for s in self.iter_slice())

    def to_vpt(self):
        """Returns a simple Radar object in VPT scan type with spectra moments
        such as reflectivity and mean velocity."""
        from ..retrieve import spectra_moments
        from ..testing import make_empty_ppi_radar
        from ..util import to_vpt

        rng_len = len(self.range.values)
        time_len = len(self.time.values)
        vpt_radar = make_empty_ppi_radar(
            ngates=rng_len, rays_per_sweep=time_len, nsweeps=1
        )

        fields = spectra_moments(self)

        rng_dict = get_metadata("range")
        rng_dict["data"] = self.range.values

        time_dict = get_metadata("time")
        time_dict["data"] = self.time.values

        vpt_radar.range = rng_dict
        vpt_radar.time = time_dict
        vpt_radar.fields = fields
        vpt_radar.metadata["instrument_name"] = "KAZR"
        to_vpt(vpt_radar)
        return vpt_radar


def _rays_per_sweep_data_factory(radar):
    """Return a function which returns the number of rays per sweep."""
    rays_per_sweep_dict = get_metadata("rays_per_sweep")
    rays_per_sweep = (
        radar.sweep_end_ray_index.values - radar.sweep_start_ray_index.values + 1
    )
    radar["rays_per_sweep"] = xr.DataArray(
        np.array(rays_per_sweep), attrs=rays_per_sweep_dict
    )


def _gate_data_factory(radar):
    """Return a function which returns the Cartesian locations of gates."""
    ranges = radar.range.values
    azimuths = radar.azimuth.values
    elevations = radar.elevation.values
    cartesian_coords = antenna_vectors_to_cartesian(
        ranges, azimuths, elevations, edges=False
    )

    # load x, y, and z data except for the coordinate in question
    gate_x_dict = get_metadata("gate_x")
    radar["gate_x"] = xr.DataArray(
        cartesian_coords[0], dims=("time", "range"), attrs=gate_x_dict
    )
    gate_y_dict = get_metadata("gate_y")
    radar["gate_y"] = xr.DataArray(
        cartesian_coords[1], dims=("time", "range"), attrs=gate_y_dict
    )
    gate_z_dict = get_metadata("gate_z")
    radar["gate_z"] = xr.DataArray(
        cartesian_coords[2], dims=("time", "range"), attrs=gate_z_dict
    )


def _gate_lon_lat_data_factory(radar):
    """Return a function which returns the geographic locations of gates."""
    x = radar.gate_x.values
    y = radar.gate_y.values
    projparams = radar.projection.copy()
    if projparams.pop("_include_lon_0_lat_0", False):
        projparams["lon_0"] = radar.longitude.values
        projparams["lat_0"] = radar.latitude.values
    geographic_coords = cartesian_to_geographic(x, y, projparams)
    # set the other geographic coordinate
    gate_latitude_dict = get_metadata("gate_latitude")
    radar["gate_latitude"] = xr.DataArray(
        geographic_coords[1], dims=("time", "range"), attrs=gate_latitude_dict
    )
    gate_longitude_dict = get_metadata("gate_longitude")
    radar["gate_longitude"] = xr.DataArray(
        geographic_coords[0], dims=("time", "range"), attrs=gate_longitude_dict
    )


def _gate_altitude_data_factory(radar):
    """Return a function which returns the gate altitudes."""
    try:
        alt = radar.altitude.values + radar.gate_z.values
    except ValueError:
        alt = np.mean(radar.altitude.values) + radar.gate_z.values
    gate_altitude_dict = get_metadata("gate_altitude")
    radar["gate_altitude"] = xr.DataArray(
        alt, dims=("time", "range"), attrs=gate_altitude_dict
    )



""" Unit Tests for Py-ART's io/mdv_common.py module. """

import warnings
from datetime import datetime
from io import BytesIO

import numpy as np
import pytest
from numpy.testing import assert_almost_equal

import pyart
from pyart.io.mdv_common import MdvFile

#################
# MdvFile tests #
#################


mdvfile = MdvFile(pyart.testing.MDV_PPI_FILE)


def test_master_header():
    # test the master header
    ref_master_header = {
        "chunk_hdr_offset": 2464,
        "data_collection_type": 0,
        "data_dimension": 0,
        "data_ordering": 0,
        "data_set_info": "MDV radar volume file created by Dsr2Vol.",
        "data_set_name": "C-SAPR",
        "data_set_source": "ARM SGP C-SAPR",
        "field_grids_differ": 0,
        "field_hdr_offset": 1024,
        "grid_orientation": 1,
        "index_number": 611,
        "native_vlevel_type": 9,
        "nchunks": 3,
        "nfields": 1,
        "max_nx": 110,
        "max_ny": 360,
        "max_nz": 1,
        "num_data_times": 1,
        "record_len1": 1016,
        "record_len2": 1016,
        "revision_number": 1,
        "sensor_alt": 0.32760000228881836,
        "sensor_lat": 36.79615783691406,
        "sensor_lon": -97.45054626464844,
        "struct_id": 14142,
        "time_begin": 1305889260,
        "time_centroid": 1305889595,
        "time_end": 1305889595,
        "time_expire": 1305890265,
        "time_gen": 1305889595,
        "time_written": 1305889668,
        "unused_fl3212": (0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0),
        "unused_si325": (0, 0, 0, 0, 0),
        "user_data": 0,
        "user_data_fl326": (0.0, 0.0, 0.0, 0.0, 0.0, 0.0),
        "user_data_si328": (0, 0, 0, 0, 0, 0, 0, 0),
        "user_time": 0,
        "vlevel_hdr_offset": 1440,
        "vlevel_included": 1,
        "vlevel_type": 9,
    }

    for k, v in ref_master_header.items():
        print("checking key:", k)
        assert mdvfile.master_header[k] == v


def test_field_header():
    # test a few of the field_header
    assert len(mdvfile.field_headers) == 1
    assert mdvfile.field_headers[0]["bad_data_value"] == 0.0
    assert mdvfile.field_headers[0]["dz_constant"] == 0
    assert mdvfile.field_headers[0]["encoding_type"] == 2
    assert mdvfile.field_headers[0]["field_name"] == "DBZ_F"
    assert mdvfile.field_headers[0]["nz"] == 1


def test_vlevel_headers():
    # test the vlevel_headers
    assert len(mdvfile.vlevel_headers) == 1
    assert mdvfile.vlevel_headers[0]["record_len1"] == 1016
    assert mdvfile.vlevel_headers[0]["record_len2"] == 1016
    assert mdvfile.vlevel_headers[0]["unused_fl32"] == (0.0, 0.0, 0.0, 0.0, 0.0)
    assert mdvfile.vlevel_headers[0]["unused_si32"] == (0, 0, 0, 0)
    assert mdvfile.vlevel_headers[0]["struct_id"] == 14144
    assert len(mdvfile.vlevel_headers[0]["level"]) == 122
    assert mdvfile.vlevel_headers[0]["level"][0] == 0.75
    assert mdvfile.vlevel_headers[0]["level"][10] == 11.699999809265137
    assert mdvfile.vlevel_headers[0]["level"][17] == 0.0
    assert len(mdvfile.vlevel_headers[0]["type"]) == 122
    assert mdvfile.vlevel_headers[0]["type"][0:17] == (9,) * 17
    assert mdvfile.vlevel_headers[0]["type"][18] == 0


def test_chunk_headers():
    # test chunk_headers
    assert len(mdvfile.chunk_headers) == 3
    ref_chunk_header_0 = {
        "chunk_data_offset": 68580,
        "chunk_id": 3,
        "info": "DsRadar params",
        "record_len1": 504,
        "record_len2": 504,
        "size": 240,
        "struct_id": 14145,
        "unused_si32": (0, 0),
    }
    for k, v in ref_chunk_header_0.items():
        print("checking key:", k)
        assert mdvfile.chunk_headers[0][k] == v


def test_calib_info():
    # test calib_info
    ref_calib_info = {
        "I0_h_co_dbm": 0.0,
        "I0_h_cx_dbm": 0.0,
        "I0_v_co_dbm": 0.0,
        "I0_v_cx_dbm": 0.0,
        "antenna_gain_h_db": 44.5,
        "antenna_gain_v_db": 44.5,
        "beamwidth_h_deg": 1.0,
        "beamwidth_v_deg": 1.0,
        "coupler_fwd_loss_h_db": 50.0,
        "coupler_fwd_loss_v_db": 50.0,
        "day": 15,
        "filter_loss_db": 1.600000023841858,
        "hour": 18,
        "ldr_h_bias_db": 0.0,
        "ldr_v_bias_db": 0.0,
        "minute": 49,
        "month": 4,
        "noise_h_co_dbm": -73.5886001586914,
        "noise_h_cx_dbm": -73.5886001586914,
        "noise_source_h_dbm": -9999.0,
        "noise_source_v_dbm": -9999.0,
        "noise_v_co_dbm": -73.13050079345703,
        "noise_v_cx_dbm": -73.13050079345703,
        "power_meas_loss_h_db": 4.96999979019165,
        "power_meas_loss_v_db": 5.840000152587891,
        "pulse_width_us": 0.800000011920929,
        "radar_constant_h_db": 73.25920104980469,
        "radar_constant_v_db": 73.25920104980469,
        "radar_name": "C-SAPR",
        "rx_gain_h_co_dbm": 33.31079864501953,
        "rx_gain_h_cx_dbm": 33.31079864501953,
        "rx_gain_v_co_dbm": 34.24689865112305,
        "rx_gain_v_cx_dbm": 34.24689865112305,
        "rx_slope_h_co_db": 1.0030399560928345,
        "rx_slope_h_cx_db": 1.0030399560928345,
        "rx_slope_v_co_db": 0.9963300228118896,
        "rx_slope_v_cx_db": 0.9963300228118896,
        "second": 36,
        "spare": (0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0),
        "sun_h_co_dbm": -9999.0,
        "sun_h_cx_dbm": -9999.0,
        "sun_v_co_dbm": -9999.0,
        "sun_v_cx_dbm": -9999.0,
        "system_phidp_deg": -60.0,
        "test_pulse_h_dbm": -9999.0,
        "test_pulse_v_dbm": -9999.0,
        "twoway_radome_loss_h_db": 0.5,
        "twoway_radome_loss_v_db": 0.5,
        "twoway_waveguide_loss_h_db": 4.0,
        "twoway_waveguide_loss_v_db": 4.0,
        "wavelength_cm": 5.329639911651611,
        "xmit_power_h_dbm": 83.6500015258789,
        "xmit_power_v_dbm": 83.6500015258789,
        "year": 2011,
        "zdr_bias_db": -1.2799999713897705,
        "zh1km_co_dbz": -33.63949966430664,
        "zh1km_cx_dbz": -33.63949966430664,
        "zv1km_co_dbz": -34.11470031738281,
        "zv1km_cx_dbz": -34.11470031738281,
    }

    for k, v in ref_calib_info.items():
        print("checking key:", k)
        assert mdvfile.calib_info[k] == v


def test_radar_info():

    # check radar_info
    ref_radar_info = {
        "altitude_km": 0.32760000228881836,
        "antenna_gain_db": 44.5,
        "field_flag": 0,
        "follow_mode": 1,
        "gate_spacing_km": 0.11991698294878006,
        "horiz_beam_width_deg": 1.0,
        "latitude_deg": 36.79615783691406,
        "longitude_deg": -97.45054626464844,
        "measXmitPowerDbmH_dbm": 82.7537612915039,
        "measXmitPowerDbmV_dbm": 82.65692138671875,
        "nfields": 1,
        "nfields_current": 0,
        "ngates": 110,
        "polarization": 5,
        "prf_hz": 1239.9993896484375,
        "prf_mode": 1,
        "prt2_s": 0.0,
        "prt_s": 0.0008064520079642534,
        "pulse_width_us": 0.800000011920929,
        "radar_constant": 73.25920104980469,
        "radar_id": 0,
        "radar_name": "C-SAPR",
        "radar_type": 0,
        "receiver_gain_db": 33.31079864501953,
        "receiver_mds_dbm": -106.89939880371094,
        "samples_per_beam": 128,
        "scan_mode": 8,
        "scan_type": 0,
        "scan_type_name": "Default_s",
        "spare_floats": (0.0, 0.0, 0.0, 0.0),
        "spare_ints": (0, 0),
        "start_range_km": 0.11787839233875275,
        "system_gain_db": 77.81079864501953,
        "unambig_range_km": 117.99627685546875,
        "unambig_vel_mps": 16.52467918395996,
        "vert_beam_width_deg": 1.0,
        "wavelength_cm": 5.330544471740723,
        "xmit_peak_pwr_watts": 231739552.0,
    }

    for k, v in ref_radar_info.items():
        print("checking key:", k)
        assert mdvfile.radar_info[k] == v


def test_geometry():
    # test geometry attributes
    az_deg, range_km, el_deg = mdvfile._calc_geometry()
    assert np.all(az_deg == np.arange(360))
    assert len(range_km) == 110
    assert_almost_equal(range_km[10], 1.32, 2)
    assert len(el_deg) == 1
    assert el_deg[0] == 0.75


def test_geometry_raises():
    mdvfile = MdvFile(pyart.testing.MDV_GRID_FILE)
    pytest.raises(NotImplementedError, mdvfile._calc_geometry)


def test_mdv_time():
    # test times attribute
    assert mdvfile.times["time_begin"] == datetime(2011, 5, 20, 11, 1)
    assert mdvfile.times["time_centroid"] == datetime(2011, 5, 20, 11, 6, 35)
    assert mdvfile.times["time_end"] == datetime(2011, 5, 20, 11, 6, 35)


def test_cart():
    # test cartography information
    assert mdvfile.projection == "ppi"
    carts = mdvfile._make_carts_dict()
    assert carts["x"].shape == (1, 360, 110)
    assert carts["y"].shape == (1, 360, 110)
    assert carts["z"].shape == (1, 360, 110)
    assert_almost_equal(carts["x"][0, 1, 2], 6.24, 2)
    assert_almost_equal(carts["y"][0, 1, 2], 357.63, 2)
    assert_almost_equal(carts["z"][0, 1, 2], 4.69, 2)


def test_fields():
    # test fields
    ref_fields = [
        "DBZ_F",
    ]
    for a, b in zip(ref_fields, mdvfile.fields):
        assert a == b


def test_fileptr():
    # test fileptr
    assert hasattr(mdvfile.fileptr, "read")


def test_read_one_field():

    mdvfile = MdvFile(pyart.testing.MDV_PPI_FILE)
    # extract a field
    assert mdvfile.fields_data[0] is None
    sweeps = mdvfile.read_a_field(0)
    assert sweeps.shape == (1, 360, 110)
    assert_almost_equal(sweeps[0, 1, 2], 13.19, 2)
    assert mdvfile.fields_data[0] is not None

    # reread, should be faster
    sweeps = mdvfile.read_a_field(0)
    assert sweeps.shape == (1, 360, 110)
    assert_almost_equal(sweeps[0, 1, 2], 13.19, 2)
    assert mdvfile.fields_data[0] is not None


def test_read_all_fields():

    mdvfile = MdvFile(pyart.testing.MDV_PPI_FILE)
    # read all fields
    assert mdvfile.fields_data[0] is None
    mdvfile.read_all_fields()
    assert mdvfile.fields_data[0] is not None
    assert_almost_equal(mdvfile.fields_data[0][0, 1, 2], 13.19, 2)


def test_read_all_fields_on_creation():
    mdvfile2 = MdvFile(pyart.testing.MDV_PPI_FILE, read_fields=True)
    assert mdvfile2.fields_data[0] is not None
    assert_almost_equal(mdvfile2.fields_data[0][0, 1, 2], 13.19, 2)
    mdvfile2.close()


def test_mdvfile_radar_stubs():
    mdvfile = pyart.io.mdv_common.MdvFile(None)
    # These methods are included in MdvFile as a stub for a future
    # write_radar_mdv function, test that the dictionaries they return have the
    # correct length.
    dic = mdvfile._get_chunk_header()
    assert len(dic) == 8
    assert isinstance(dic, dict)

    dic = mdvfile._get_radar_info()
    assert len(dic) == 38
    assert isinstance(dic, dict)

    dic = mdvfile._get_calib()
    assert len(dic) == 59
    assert isinstance(dic, dict)

    dic = mdvfile._get_compression_info()
    assert len(dic) == 5
    assert isinstance(dic, dict)

    dic = mdvfile._get_levels_info(10)
    assert len(dic) == 2
    assert isinstance(dic, dict)


def test_rhi_cart():
    mdvfile = pyart.io.mdv_common.MdvFile(pyart.testing.MDV_RHI_FILE)
    assert mdvfile.projection == "rhi"
    carts = mdvfile._make_carts_dict()
    assert carts["x"].shape == (1, 283, 125)
    assert carts["y"].shape == (1, 283, 125)
    assert carts["z"].shape == (1, 283, 125)
    assert_almost_equal(carts["x"][0, 1, 2], -52.63, 2)
    assert_almost_equal(carts["y"][0, 1, 2], -332.31, 2)
    assert_almost_equal(carts["z"][0, 1, 2], 121.47, 2)


class Mdv_common_Tests:
    """
    Class for declaring unit tests for the io/mdv_common.py module which
    require a temporary file which will be removed at the end of the test.
    """

    def test_mdv_file_read_write_radar(self):
        with warnings.catch_warnings(record=True) as w:
            mdvfile_orig = pyart.io.mdv_common.MdvFile(pyart.testing.MDV_PPI_FILE)
            mdvfile_orig.read_all_fields()

            inmemfile = BytesIO()
            mdvfile_orig.write(inmemfile)
            inmemfile.seek(0)
            # check that a UserWarning was issued since zlib compression used
            assert len(w) == 1
            assert issubclass(w[-1].category, UserWarning)

            mdvfile = pyart.io.mdv_common.MdvFile(inmemfile)
            self.check_mdvfile_ppi(mdvfile)

    def test_read_write_file_objects(self):
        # read from and write two a file object
        f = open(pyart.testing.MDV_PPI_FILE, "rb")
        mdvfile = pyart.io.mdv_common.MdvFile(f)
        mdvfile.read_all_fields()
        self.check_mdvfile_ppi(mdvfile)

        # write out the file using an file handler
        f2 = BytesIO()
        with warnings.catch_warnings():
            # ignore the UserWarning about the non-implement compression_type
            warnings.simplefilter("ignore", category=UserWarning)
            mdvfile.write(f2)
        f.close()
        f2.seek(0)

        # re-read and check object
        mdvfile = pyart.io.mdv_common.MdvFile(f2)
        self.check_mdvfile_ppi(mdvfile)

    @staticmethod
    def check_mdvfile_ppi(mdvfile):
        # check some parameters
        assert mdvfile.master_header["time_end"] == 1305889595
        assert mdvfile.master_header["chunk_hdr_offset"] == 2464
        assert len(mdvfile.field_headers) == 1
        assert mdvfile.field_headers[0]["bad_data_value"] == 0.0
        assert len(mdvfile.vlevel_headers) == 1
        assert mdvfile.vlevel_headers[0]["record_len1"] == 1016
        assert mdvfile.chunk_headers[0]["chunk_id"] == 3
        assert mdvfile.calib_info["day"] == 15
        assert mdvfile.radar_info["scan_mode"] == 8
        az_deg, range_km, el_deg = mdvfile._calc_geometry()
        assert np.all(az_deg == np.arange(360))
        assert len(range_km) == 110
        assert_almost_equal(range_km[10], 1.32, 2)
        assert len(el_deg) == 1
        assert el_deg[0] == 0.75
        assert mdvfile.times["time_begin"] == datetime(2011, 5, 20, 11, 1)
        assert len(mdvfile.elevations) == 0

    def test_mdv_file_read_write_radar_rhi(self):
        # write and read the RHI file which contains a elevation chunk
        mdvfile = pyart.io.mdv_common.MdvFile(pyart.testing.MDV_RHI_FILE)
        mdvfile.read_all_fields()
        inmemfile = BytesIO()
        with warnings.catch_warnings():
            # ignore the UserWarning about the non-implement compression_type
            warnings.simplefilter("ignore", category=UserWarning)
            mdvfile.write(inmemfile)
        inmemfile.seek(0)
        mdvfile2 = pyart.io.mdv_common.MdvFile(inmemfile)

        # verify that the elevations are similar
        assert len(mdvfile2.elevations) == len(mdvfile.elevations)
        assert mdvfile2.elevations[0] == mdvfile.elevations[0]



"""
Utilities helpful when writing and running unit tests such as sample files
and sample objects.

"""

from .example_data import get_test_data  # noqa
from .sample_files import *  # noqa
from .sample_objects import *  # noqa
from .tmpdirs import InTemporaryDirectory  # noqa

__all__ = [s for s in dir() if not s.startswith("_")]



""" Unit tests for bias and noise module. """

import pytest
from numpy.testing import assert_allclose

import pyart

radar = pyart.io.read(pyart.testing.NEXRAD_ARCHIVE_MSG31_FILE)


def test_correct_bias():
    corr_field_expected = [-32.0, -32.0, -32.0]
    corr_test = pyart.correct.correct_bias(radar, field_name=None)

    assert_allclose(corr_test["data"][0][0:3], corr_field_expected, atol=1e-14)
    assert_allclose(corr_test["data"][-1][0:3], corr_field_expected, atol=1e-14)
    assert corr_test["long_name"] == "Corrected reflectivity"

    radar.add_field("corrected_reflectivity", corr_test, replace_existing=True)

    corr_test_2 = pyart.correct.correct_bias(radar, field_name="corrected_reflectivity")
    assert corr_test_2["long_name"] == "Corrected reflectivity"

    bias = 0
    field_name = "foo"
    pytest.raises(KeyError, pyart.correct.correct_bias, radar, bias, field_name)



"""
Utilities for reading gamic hdf5 files.

"""

# TODO to move out of aux_io namespace:
# * unit tests - need small sample files.
# * auto-detect file type with pyart.io.read function
# * move to pyart.io namespace

import datetime
import warnings

import numpy as np

from ..config import FileMetadata, get_fillvalue
from ..core.radar import Radar
from ..io.common import _test_arguments, make_time_unit_str

try:
    from .gamicfile import GAMICFile

    _H5PY_AVAILABLE = True
except ImportError:
    _H5PY_AVAILABLE = False
from ..exceptions import MissingOptionalDependency

LIGHT_SPEED = 2.99792458e8  # speed of light in meters per second


def read_gamic(
    filename,
    field_names=None,
    additional_metadata=None,
    file_field_names=False,
    exclude_fields=None,
    include_fields=None,
    valid_range_from_file=True,
    units_from_file=True,
    pulse_width=None,
    **kwargs
):
    """
    Read a GAMIC hdf5 file.

    Parameters
    ----------
    filename : str
        Name of GAMIC HDF5 file to read data from.
    field_names : dict, optional
        Dictionary mapping field names in the file names to radar field names.
        Unlike other read functions, fields not in this dictionary or having a
        value of None are still included in the radar.fields dictionary, to
        exclude them use the `exclude_fields` parameter. Fields which are
        mapped by this dictionary will be renamed from key to value.
    additional_metadata : dict of dicts, optional
        This parameter is not used, it is included for uniformity.
    file_field_names : bool, optional
        True to force the use of the field names from the file in which
        case the `field_names` parameter is ignored. False will use to
        `field_names` parameter to rename fields.
    exclude_fields : list or None, optional
        List of fields to exclude from the radar object. This is applied
        after the `file_field_names` and `field_names` parameters. Set
        to None to include all fields specified by include_fields.
    include_fields : list or None, optional
        List of fields to include from the radar object. This is applied
        after the `file_field_names` and `field_names` parameters. Set
        to None to include all fields not specified by exclude_fields.
    valid_range_from_file : bool, optional
        True to extract valid range (valid_min and valid_max) for all
        field from the file when they are present.  False will not extract
        these parameters.
    units_from_file : bool, optional
        True to extract the units for all fields from the file when available.
        False will not extract units using the default units for the fields.
    pulse_width : list or None,
        Mandatory for gamic radar processors which have pulsewidth enums.
        pulse_width should contain the pulsewidth' in us.

    Returns
    -------
    radar : Radar
        Radar object.

    """
    # check that h5py is available
    if not _H5PY_AVAILABLE:
        raise MissingOptionalDependency(
            "h5py is required to use read_gamic but is not installed"
        )

    # test for non empty kwargs
    _test_arguments(kwargs)

    # create metadata retrieval object
    filemetadata = FileMetadata(
        "gamic",
        field_names,
        additional_metadata,
        file_field_names,
        exclude_fields,
        include_fields,
    )

    # Open HDF5 file and get handle
    gfile = GAMICFile(filename)

    # verify that all scans are present in file
    assert gfile.is_file_complete()

    # latitude, longitude and altitude
    latitude = filemetadata("latitude")
    longitude = filemetadata("longitude")
    altitude = filemetadata("altitude")
    latitude["data"] = gfile.where_attr("lat", "float64")
    longitude["data"] = gfile.where_attr("lon", "float64")
    altitude["data"] = gfile.where_attr("height", "float64")

    # metadata
    metadata = filemetadata("metadata")
    metadata["original_container"] = "GAMIC-HDF5"
    what_mapping = {
        "version": "gamic_version",
        "sets_scheduled": "sets_scheduled",
        "object": "gamic_object",
        "date": "gamic_date",
    }
    for gamic_key, metadata_key in what_mapping.items():
        if gfile.is_attr_in_group("what", gamic_key):
            metadata[metadata_key] = gfile.raw_group_attr("what", gamic_key)

    how_keys = [
        "software",
        "template_name",
        "site_name",
        "host_name",
        "azimuth_beam",
        "elevation_beam",
        "sdp_name",
        "sw_version",
        "sdp_version",
        "simulated",
    ]
    for key in how_keys:
        if gfile.is_attr_in_group("how", key):
            metadata[key] = gfile.raw_group_attr("how", key)

    # sweep_start_ray_index, sweep_end_ray_index
    sweep_start_ray_index = filemetadata("sweep_start_ray_index")
    sweep_end_ray_index = filemetadata("sweep_end_ray_index")
    sweep_start_ray_index["data"] = gfile.start_ray.astype("int32")
    sweep_end_ray_index["data"] = gfile.end_ray.astype("int32")

    # sweep number
    sweep_number = filemetadata("sweep_number")
    try:
        sweep_number["data"] = gfile.what_attrs("set_idx", "int32")
    except KeyError:
        sweep_number["data"] = np.arange(gfile.nsweeps, dtype="int32")

    # sweep_type
    scan_type = gfile.raw_scan0_group_attr("what", "scan_type").lower()
    if hasattr(scan_type, "decode"):
        scan_type = scan_type.decode("utf-8")
    # check that all scans in the volume are the same type
    if not gfile.is_file_single_scan_type():
        raise NotImplementedError("Mixed scan_type volume.")
    if scan_type not in ["ppi", "rhi"]:
        message = "Unknown scan type: %s, reading as RHI scans." % (scan_type)
        warnings.warn(message)
        scan_type = "rhi"

    # sweep_mode, fixed_angle
    sweep_mode = filemetadata("sweep_mode")
    fixed_angle = filemetadata("fixed_angle")
    if scan_type == "rhi":
        sweep_mode["data"] = np.array(gfile.nsweeps * ["rhi"])
        fixed_angle["data"] = gfile.how_attrs("azimuth", "float32")
    elif scan_type == "ppi":
        sweep_mode["data"] = np.array(gfile.nsweeps * ["azimuth_surveillance"])
        fixed_angle["data"] = gfile.how_attrs("elevation", "float32")

    # time
    time = filemetadata("time")
    t_data = gfile.ray_header("timestamp", "int64")
    start_epoch = t_data[0] // 1.0e6  # truncate to second resolution
    start_time = datetime.datetime.utcfromtimestamp(start_epoch)
    time["units"] = make_time_unit_str(start_time)
    time["data"] = ((t_data - start_epoch * 1.0e6) / 1.0e6).astype("float64")

    # range
    _range = filemetadata("range")
    ngates = gfile.max_num_gates
    range_start = float(gfile.raw_scan0_group_attr("how", "range_start"))
    # n_samples insertion
    range_samples = int(gfile.raw_scan0_group_attr("how", "range_samples"))
    range_step = float(gfile.raw_scan0_group_attr("how", "range_step")) * range_samples
    # range_step may need to be scaled by range_samples
    # XXX This gives distances to start of gates not center, this matches
    # Radx but may be incorrect, add range_step / 2. for center
    _range["data"] = np.arange(ngates, dtype="float32") * range_step + range_start
    _range["meters_to_center_of_first_gate"] = range_start
    _range["meters_between_gates"] = range_step

    # elevation
    elevation = filemetadata("elevation")
    start_angle = gfile.ray_header("elevation_start", "float32")
    stop_angle = gfile.ray_header("elevation_stop", "float32")
    elevation["data"] = _avg_radial_angles(start_angle, stop_angle)

    # azimuth
    azimuth = filemetadata("azimuth")
    start_angle = gfile.ray_header("azimuth_start", "float32")
    stop_angle = gfile.ray_header("azimuth_stop", "float32")
    azimuth["data"] = _avg_radial_angles(start_angle, stop_angle) % 360.0

    # fields
    fields = {}
    moment_groups = gfile.moment_groups()
    moment_names = gfile.moment_names(moment_groups)

    for moment_name, group in zip(moment_names, moment_groups):

        field_name = filemetadata.get_field_name(moment_name)
        if field_name is None:
            continue

        field_dic = filemetadata(field_name)
        field_dic["data"] = gfile.moment_data(group, "float32")
        field_dic["_FillValue"] = get_fillvalue()

        if valid_range_from_file:
            try:
                valid_min = gfile.raw_scan0_group_attr(group, "dyn_range_min")
                valid_max = gfile.raw_scan0_group_attr(group, "dyn_range_max")
                field_dic["valid_min"] = valid_min.decode("utf-8")
                field_dic["valid_max"] = valid_max.decode("utf-8")
            except:
                pass

        if units_from_file:
            try:
                units = gfile.raw_scan0_group_attr(group, "unit")
                field_dic["units"] = units.decode("utf-8")
            except:
                pass
        fields[field_name] = field_dic

    # ray_angle_res
    ray_angle_res = filemetadata("ray_angle_res")
    ray_angle_res["data"] = gfile.how_attrs("angle_step", "float32")

    # rays_are_indexed
    rays_are_indexed = filemetadata("rays_are_indexed")
    rays_are_indexed["data"] = np.array(
        [["false", "true"][i] for i in gfile.how_attrs("angle_sync", "uint8")]
    )

    # target_scan_rate
    target_scan_rate = filemetadata("target_scan_rate")
    target_scan_rate["data"] = gfile.how_attrs("scan_speed", "float32")

    # scan_rate
    scan_rate = filemetadata("scan_rate")
    scan_rate["data"] = target_scan_rate["data"]
    if scan_type == "ppi":
        azs_names = ["az_speed", "azimuth_speed"]
        azs_name = azs_names[0]
        for azs_name in azs_names:
            if gfile.is_field_in_ray_header(azs_name):
                scan_rate["data"] = gfile.ray_header(azs_name, "float32")
                break
    elif scan_type == "rhi":
        els_names = ["el_speed", "elevation_speed"]
        els_name = els_names[0]
        for els_name in els_names:
            if gfile.is_field_in_ray_header(els_name):
                scan_rate["data"] = gfile.ray_header(els_name, "float32")
                break
    else:
        scan_rate = None

    # instrument_parameters
    instrument_parameters = _get_instrument_params(gfile, filemetadata, pulse_width)

    gfile.close()

    return Radar(
        time,
        _range,
        fields,
        metadata,
        scan_type,
        latitude,
        longitude,
        altitude,
        sweep_number,
        sweep_mode,
        fixed_angle,
        sweep_start_ray_index,
        sweep_end_ray_index,
        azimuth,
        elevation,
        instrument_parameters=instrument_parameters,
        ray_angle_res=ray_angle_res,
        rays_are_indexed=rays_are_indexed,
        scan_rate=scan_rate,
        target_scan_rate=target_scan_rate,
    )


def _get_instrument_params(gfile, filemetadata, pulse_width):
    """Return a dictionary containing instrument parameters."""

    instrument_params = {}

    dic = filemetadata("frequency")
    dic["data"] = np.array(
        [LIGHT_SPEED / gfile.raw_scan0_group_attr("how", "radar_wave_length")],
        dtype="float32",
    )
    instrument_params["frequency"] = dic

    dic = filemetadata("radar_beam_width_h")
    dic["data"] = gfile.how_attr("azimuth_beam", "float32")
    instrument_params["radar_beam_width_h"] = dic

    dic = filemetadata("radar_beam_width_v")
    dic["data"] = gfile.how_attr("elevation_beam", "float32")
    instrument_params["radar_beam_width_v"] = dic

    dic = filemetadata("pulse_width")
    pw_names = ["pulse_width_us", "pulse_width_mks", "pulse_width"]
    pw_name = "pulse_width_us"
    dic["data"] = None
    for pw_name in pw_names:
        if gfile.is_attr_in_group("/scan0/how", pw_name):
            if pw_name == "pulse_width":
                dic["data"] = gfile.sweep_expand(
                    pulse_width[gfile.how_attrs(pw_name, "int")[0]] * 1e-6
                )
            else:
                dic["data"] = gfile.sweep_expand(
                    gfile.how_attrs(pw_name, "float32") * 1e-6
                )
            break
    instrument_params["pulse_width"] = dic

    dic = filemetadata("prt")
    dic["data"] = gfile.sweep_expand(1.0 / gfile.how_attrs("PRF", "float32"))
    instrument_params["prt"] = dic

    unfolding = gfile.how_attrs("unfolding", "int32")
    dic = filemetadata("prt_mode")
    dic["data"] = np.array([_prt_mode_from_unfolding(i) for i in unfolding])
    instrument_params["prt_mode"] = dic

    dic = filemetadata("prt_ratio")
    dic["data"] = gfile.sweep_expand(
        [[1, 2.0 / 3.0, 3.0 / 4.0, 4.0 / 5.0][i] for i in unfolding]
    )
    instrument_params["prt_ratio"] = dic

    dic = filemetadata("unambiguous_range")
    dic["data"] = gfile.sweep_expand(gfile.how_attrs("range", "float32"))
    instrument_params["unambiguous_range"] = dic

    if gfile.is_attr_in_group("/scan0/how/extended", "nyquist_velocity"):
        dic = filemetadata("nyquist_velocity")
        dic["data"] = gfile.sweep_expand(gfile.how_ext_attrs("nyquist_velocity"))
        instrument_params["nyquist_velocity"] = dic

    dic = filemetadata("n_samples")
    dic["data"] = gfile.sweep_expand(
        gfile.how_attrs("range_samples", "int32")
        * gfile.how_attrs("time_samples", "int32"),
        dtype="int32",
    )
    instrument_params["n_samples"] = dic

    return instrument_params


def _avg_radial_angles(angle1, angle2):
    """Return the average angle between two radial angles."""
    return np.angle(
        (np.exp(1.0j * np.deg2rad(angle1)) + np.exp(1.0j * np.deg2rad(angle2))) / 2.0,
        deg=True,
    )


def _prt_mode_from_unfolding(unfolding):
    """Return 'fixed' or 'staggered' depending on unfolding flag"""
    if unfolding == 0:
        return "fixed"
    else:
        return "staggered"



"""
A script for comparing the e Vulpiani, Maesaka and Schneedbeli methods
for KDP estimation with sample data.
"""

import matplotlib.pyplot as plt
import numpy as np
import pytest

from pyart.config import get_field_name
from pyart.retrieve.kdp_proc import kdp_maesaka, kdp_schneebeli, kdp_vulpiani
from pyart.testing import sample_objects


def _make_real_psidp_radar():
    """
    Create single-ray radar with linear differential phase profile with
    specified slope.
    ---
    Returns
    -------
    radar : Radar
            PyART radar instance with differential phase profile in deg.
    """
    psidp = np.array(
        [
            [
                -2.33313751e00,
                1.80617523e00,
                7.17742920e-01,
                1.82811661e01,
                1.89352417e01,
                1.67904205e01,
            ]
        ]
    )
    psidp = np.ma.array(psidp)
    radar = sample_objects.make_empty_ppi_radar(len(psidp[0]), 1, 1)
    psidp_dict = {
        "data": psidp,
    }
    radar.add_field(get_field_name("differential_phase"), psidp_dict)
    # Define real ranges
    radar.range["data"] = 75 * np.arange(0, len(psidp[0]))
    return radar


@pytest.mark.mpl_image_compare(tolerance=30)
def test_compare_kdp_estimation_methods():
    # Get profile of noisy psidp
    prof_psidp = _make_real_psidp_radar()
    # Maesaka method
    kdp_mae, phidpf_mae, phidp_mae = kdp_maesaka(
        prof_psidp, maxiter=1000, check_outliers=False, parallel=False
    )

    # Vulpiani method (note windsize is just a guess here..)
    kdp_vulp, phidp_vulp = kdp_vulpiani(
        prof_psidp, windsize=2, n_iter=20, parallel=False, band="X"
    )
    # Kalman filter method
    kdp_schnee, kdp_std_schnee, phidp_schnee = kdp_schneebeli(
        prof_psidp, parallel=False, band="X"
    )
    # Create figure
    fig = plt.figure(figsize=(10, 10))
    plt.subplot(2, 1, 1)
    plt.grid(True)
    plt.title("Kdp estimation")
    ranges = prof_psidp.range["data"]
    plt.plot(ranges, kdp_mae["data"][0])
    plt.plot(ranges, kdp_vulp["data"][0])
    plt.plot(ranges, kdp_schnee["data"][0])
    plt.xlabel("Range [m]")
    plt.ylabel("Kdp [deg/km]")
    plt.legend(["Maesaka", "Vulpiani", "Schneebeli"], loc=0)
    plt.subplot(2, 1, 2)
    plt.grid(True)
    plt.title("Reconstructed Phidp")
    ranges = prof_psidp.range["data"]
    phidp_mae = 0.5 * (phidp_mae["data"][0] + phidp_mae["data"][0])
    plt.plot(ranges, phidp_mae)
    plt.plot(ranges, phidp_vulp["data"][0])
    plt.plot(ranges, phidp_schnee["data"][0])
    plt.plot(ranges, prof_psidp.fields["differential_phase"]["data"][0])
    plt.xlabel("Range [m]")
    plt.ylabel("Diff. phase [deg]")
    plt.legend(["Maesaka", "Vulpiani", "Schneebeli", "Real Psidp"], loc=0)
    try:
        return fig
    finally:
        plt.close(fig)



import matplotlib.pyplot as plt
import numpy as np


def plot_convstrat_scheme(
    always_core_thres,
    use_cosine,
    max_diff=None,
    zero_diff_cos_val=None,
    use_addition=False,
    scalar_diff=None,
):
    """
    Plots the scheme used in the convective stratiform classification

    Parameters
    ----------
     always_core_thres : float
        All values above this threshold considered to be convective
    use_cosine : bool
        Boolean used to determine if cosine scheme should be used for identifying convective cores (True) or a scalar
        scheme (False)
    max_diff : float, optional
        Maximum difference between background average and reflectivity in order to be classified as convective.
        "a" value in Eqn. B1 in Yuter and Houze (1997)
    zero_diff_cos_val : float, optional
        Value where difference between background average and reflectivity is zero in the cosine function
        "b" value in Eqn. B1 in Yuter and Houze (1997)
    use_addition : bool, optional
        Determines if a multiplier (False) or addition (True) in the scalar difference scheme should be used
    scalar_diff : float, optional
        If using a scalar difference scheme, this value is the multiplier or addition to the background average

    """

    # create array of background values
    bkg_vals = np.linspace(0, 60, 100)

    # create difference array
    if use_cosine:
        # cosine scheme
        diff = max_diff * np.cos(np.pi * bkg_vals / (2 * zero_diff_cos_val))
    else:
        if use_addition:
            # scalar addition scheme
            diff = (bkg_vals + scalar_diff) - bkg_vals
        else:
            # scalar multiplier scheme
            diff = (bkg_vals * scalar_diff) - bkg_vals

    # if values are less than zero, set to zero
    diff[diff < 0] = 0
    # background values greater than always core thres, set to zero
    diff[bkg_vals > always_core_thres] = 0

    # Now plot
    ax = plt.gca()
    # plot difference line
    ax.plot(bkg_vals, diff, lw=2, color="black")
    # plot always core thres
    ax.axvline(x=always_core_thres, lw=1, ls="--", color="red")
    ax.text(always_core_thres + 2, 1, "Always Core Thres.", color="red")
    if use_cosine:
        # plot zero difference cosine value
        ax.axvline(x=zero_diff_cos_val, lw=1, ls="--", color="green")
        ax.text(zero_diff_cos_val + 2, 0.75, "Zero Diff. Cosine Val.", color="green")
        # plot max difference
        ax.axhline(y=max_diff, lw=1, ls="--", color="blue")
        ax.text(10, max_diff + 0.05, "Max. Diff.", color="blue")
    elif use_addition:
        # plot scalar
        ax.axhline(y=scalar_diff, lw=1, ls="--", color="orange")
        ax.text(10, scalar_diff + 0.05, "Scalar Diff.", color="orange")
    # add grid
    ax.grid()
    # set axis limits
    ax.set_ylim([0, np.max(diff) + 0.2])
    ax.set_xlim([np.min(bkg_vals), np.max(bkg_vals)])
    # set axis labels and title
    ax.set_ylabel("Difference (dBZ - dBZ$_{background}$)")
    ax.set_xlabel("Background Value (dBZ$_{background}$)")
    ax.set_title("Convective Stratiform Equation")



"""
Utilities for reading CSU-CHILL CHL files.

"""

import struct
from datetime import datetime

import numpy as np

from ..config import FileMetadata, get_fillvalue
from ..core.radar import Radar
from .common import _test_arguments, make_time_unit_str, prepare_for_read


def read_chl(
    filename,
    field_names=None,
    additional_metadata=None,
    file_field_names=None,
    exclude_fields=None,
    include_fields=None,
    use_file_field_attributes=True,
    **kwargs
):
    """
    Read a CSU-CHILL CHL file.

    Parameters
    ----------
    filename : str
        Name of CHL file.
    field_names : dict, optional
        Dictionary mapping CHL field names to radar field names. If a
        data type found in the file does not appear in this dictionary or has
        a value of None it will not be placed in the radar.fields dictionary.
        A value of None, the default, will use the mapping defined in the
        Py-ART configuration file.
    additional_metadata : dict of dicts, optional
        Dictionary of dictionaries to retrieve metadata from during this read.
        This metadata is not used during any successive file reads unless
        explicitly included. A value of None, the default, will not
        introduct any addition metadata and the file specific or default
        metadata as specified by the Py-ART configuration file will be used.
    file_field_names : bool, optional
        True to use the CHL field names for the field names in the radar
        object. If this case the field_names parameter is ignored.
        The field dictionary will likely only have a 'data' key, unless
        the fields are defined in `additional_metadata`.
    exclude_fields : list or None, optional
        List of fields to exclude from the radar object. This is applied
        after the `file_field_names` and `field_names` parameters.
    include_fields : list or None, optional
        List of fields to include from the radar object. This is applied
        after the `field_file_names` and `field_names` parameters. Set to
        None to include all fields not in exclude_fields.
    use_file_field_attributes : bool, optional
        True to use information provided by in the file to set the field
        attribute `long_name`, `units`, `valid_max`, and `valid_min`. False
        will not set these unless they are defined in the configuration file
        or in `additional_metadata`.

    Returns
    -------
    radar : Radar
        Radar object containing data from CHL file.

    """
    # test for non empty kwargs
    _test_arguments(kwargs)

    # create metadata retrival object
    filemetadata = FileMetadata(
        "chl",
        field_names,
        additional_metadata,
        file_field_names,
        exclude_fields,
        include_fields,
    )

    # read data
    chl_file = ChlFile(prepare_for_read(filename))

    # time
    time = filemetadata("time")
    tdata = np.array(chl_file.time)
    min_time = np.floor(tdata.min())
    time["data"] = (tdata - min_time).astype("float64")
    time["units"] = make_time_unit_str(datetime.utcfromtimestamp(min_time))

    # range
    _range = filemetadata("range")
    _range["data"] = (
        np.array(range(chl_file.ngates)) * chl_file.gate_spacing
        + chl_file.first_gate_offset
    )
    _range["meters_between_gates"] = np.array(chl_file.gate_spacing)
    _range["meters_to_center_of_first_gate"] = 0.0

    # scan_type
    scan_type = SCAN_MODE_NAMES[chl_file.scan_types[-1]]

    # fields
    fields = {}
    for i, fdata in chl_file.fields.items():

        field_info = chl_file.field_info[i]
        field_name = filemetadata.get_field_name(field_info["name"])
        if field_name is None:
            continue

        field_dic = filemetadata(field_name)
        np.ma.set_fill_value(fdata, get_fillvalue())
        field_dic["data"] = fdata
        field_dic["_FillValue"] = get_fillvalue()

        if use_file_field_attributes:
            field_dic["long_name"] = field_info["descr"]
            field_dic["units"] = field_info["units"]
            field_dic["valid_max"] = field_info["max_val"]
            field_dic["valid_min"] = field_info["min_val"]

        fields[field_name] = field_dic

    # metadata
    metadata = filemetadata("metadata")
    metadata["instrument_name"] = chl_file.radar_info["radar_name"]
    metadata["original_container"] = "CHL"

    # longitude, latitude, altitude
    latitude = filemetadata("latitude")
    longitude = filemetadata("longitude")
    altitude = filemetadata("altitude")
    latitude["data"] = np.array([chl_file.radar_info["latitude"]], "f8")
    longitude["data"] = np.array([chl_file.radar_info["longitude"]], "f8")
    altitude["data"] = np.array([chl_file.radar_info["altitude"]], "f8")

    # sweep_number, sweep_mode, fixed_angle, sweep_start_ray_index,
    # sweep_end_ray_index
    sweep_number = filemetadata("sweep_number")
    sweep_mode = filemetadata("sweep_mode")
    fixed_angle = filemetadata("fixed_angle")
    sweep_start_ray_index = filemetadata("sweep_start_ray_index")
    sweep_end_ray_index = filemetadata("sweep_end_ray_index")
    sweep_number["data"] = np.arange(chl_file.num_sweeps, dtype="int32")
    sweep_mode["data"] = np.array(
        [SCAN_MODE_NAMES[i] for i in chl_file.scan_types], dtype="S"
    )
    fixed_angle["data"] = np.array(chl_file.fixed_angle, dtype="float32")

    ray_count = chl_file.rays_per_sweep
    ssri = np.cumsum(np.append([0], ray_count[:-1])).astype("int32")
    sweep_start_ray_index["data"] = ssri
    sweep_end_ray_index["data"] = np.cumsum(ray_count).astype("int32") - 1

    # azimuth, elevation
    azimuth = filemetadata("azimuth")
    elevation = filemetadata("elevation")
    azimuth["data"] = np.array(chl_file.azimuth, dtype="float32")
    elevation["data"] = np.array(chl_file.elevation, dtype="float32")

    # instrument parameters
    instrument_parameters = None

    chl_file.close()
    return Radar(
        time,
        _range,
        fields,
        metadata,
        scan_type,
        latitude,
        longitude,
        altitude,
        sweep_number,
        sweep_mode,
        fixed_angle,
        sweep_start_ray_index,
        sweep_end_ray_index,
        azimuth,
        elevation,
        instrument_parameters=instrument_parameters,
    )


class ChlFile:
    """
    A file object for CHL data.

    Parameters
    ----------
    filename : str or file-like.
        Name of CHL file to read or a file-like object pointing to the
        beginning of such a file.
    ns_time : bool
        True to determine ray collection times to the nano-second, False
        will only determine times to the second.
    debug : bool
        True to keep packet data in the _packets attribute to aid in
        debugging.

    Attributes
    ----------
    ngates : int
        Number of gates per ray.
    num_sweeps : int
        Number of sweeps in the volume.
    gate_spacing : float
        Spacing in meters between gates.
    first_gate_offset : float
        Distance in meters to the first range gate.
    time : list of ints
        Time in seconds in epoch for each ray in the volume.
    azimuth : list of floats
        Azimuth angle for each ray in the volume in degrees.
    elevation : list of floats
        Elevation angle for each ray in the volume in degrees.
    fixed_angle : list of floats
        Fixed angles for each sweep.
    sweep_number : list of ints
        Sweep numbers reported in file.
    scan_types : list of ints
        Chill defined scan type for each sweep.
    rays_per_sweep : list of ints
        Number of rays in each sweep.
    fields : dict
        Dictionary of field data index by field number.
    radar_info : dict
        Radar information recorded in the file.
    field_info : dict
        Field information (limits, name, etc.) recorded in the file.
    processor_info : dict
        Processor information recorded in the file.

    """

    def __init__(self, filename, ns_time=True, debug=False):

        # initalize attributes
        self.ngates = None
        self.num_sweeps = None
        self.gate_spacing = None
        self.time = []
        self.azimuth = []
        self.elevation = []
        self.fixed_angle = []
        self.sweep_number = []
        self.scan_types = []
        self.rays_per_sweep = None
        self.fields = {}
        self.radar_info = None
        self.field_info = {}
        self.processor_info = None
        self.first_gate_offset = None

        # private attributes
        self._dstring = b""  # string containing field data.
        self._bit_mask = None  # bit mask specifying fields present in file.
        self._dtype = None  # NumPy dtype for a single gate (all fields).
        self._ray_bsize = None  # size in bytes of a single ray (all fields).
        self._packets = []  # List of packets, not set if debug is False.
        self._field_nums = None  # Field number available in file.
        self._rays_in_current_sweep = None  # accumulator for counting rays.
        self._fh = None  # file handler
        self._include_ns_time = ns_time

        # read all blocks from the file
        if hasattr(filename, "read"):
            self._fh = filename
        else:
            self._fh = open(filename, "rb")
        packet = 1
        while packet is not None:
            packet = self._read_block()
            if debug:
                self._packets.append(packet)

        self._extract_fields()
        self.rays_per_sweep.append(self._rays_in_current_sweep)

    def close(self):
        """Close the file."""
        self._fh.close()

    def _read_block(self):
        """Read a block from an open CHL file."""
        pld = self._fh.read(8)
        if pld == b"":
            return None
        block_id, length = struct.unpack("<2i", pld)
        payload = self._fh.read(length - 8)

        # parse the block
        if block_id == ARCH_ID_FILE_HDR:
            packet = self._parse_file_hdr_block(payload)
        elif block_id == ARCH_ID_FIELD_SCALE:
            packet = self._parse_field_scale_block(payload)
        elif block_id == ARCH_ID_RAY_HDR:
            packet = self._parse_ray_hdr_block(payload)
        elif block_id == HSK_ID_RADAR_INFO:
            packet = self._parse_radar_info_block(payload)
        elif block_id == HSK_ID_PROCESSOR_INFO:
            packet = self._parse_processor_info_block(payload)
        elif block_id == HSK_ID_SCAN_SEG:
            packet = self._parse_scan_seg_block(payload)
        elif block_id == ARCH_ID_SWEEP_BLOCK:
            packet = self._parse_sweep_block(payload)
        else:
            packet = {}
        packet["block_id"] = block_id
        packet["length"] = length
        return packet

    # Block parsers
    def _parse_file_hdr_block(self, payload):
        """Parse a field_hdr block."""
        return _unpack_structure(payload, ARCH_FILE_HDR_T)

    def _parse_field_scale_block(self, payload):
        """Parse a field_scale block. Add scale to field_info attr."""
        packet = _unpack_structure(payload, FIELD_SCALE_T)
        packet["name"] = packet["name"].decode("utf-8").rstrip("\x00")
        packet["units"] = packet["units"].decode("utf-8").rstrip("\x00")
        packet["descr"] = packet["descr"].decode("utf-8").rstrip("\x00")
        self.field_info[packet["bit_mask_pos"]] = packet
        return packet

    def _parse_radar_info_block(self, payload):
        """Parse a radar_info block. Update metadata attribute."""
        packet = _unpack_structure(payload, RADAR_INFO_T)
        packet["radar_name"] = packet["radar_name"].decode("utf-8").rstrip("\x00")
        self.radar_info = packet.copy()
        return packet

    def _parse_processor_info_block(self, payload):
        """Parse a processor_info block. Set dr attribute."""
        packet = _unpack_structure(payload, PROCESSOR_INFO)
        self.gate_spacing = packet["gate_spacing"]
        self.first_gate_offset = packet["range_offset"]
        self.processor_info = packet.copy()
        return packet

    def _parse_scan_seg_block(self, payload):
        """Parse a scan_seg_block. Update sweep attributes."""
        packet = _unpack_structure(payload, SCAN_SEG)
        self.sweep_number.append(packet["sweep_num"])
        self.fixed_angle.append(packet["current_fixed_angle"])
        self.scan_types.append(packet["scan_type"])
        if self.rays_per_sweep is None:
            self.rays_per_sweep = []  # first sweep block
            self._rays_in_current_sweep = 0
        else:
            self.rays_per_sweep.append(self._rays_in_current_sweep)
            self._rays_in_current_sweep = 0
        return packet

    def _parse_sweep_block(self, payload):
        """Parse a sweep block. Set num_sweeps attribute."""
        packet = {}
        packet["num_sweeps"] = struct.unpack("I", payload[0:4])[0]
        packet["swp_offsets"] = struct.unpack(
            str(packet["num_sweeps"]) + "Q", payload[4:]
        )
        self.num_sweeps = packet["num_sweeps"]
        return packet

    def _parse_ray_hdr_block(self, payload):
        """Parse a ray_hdr block. Update associated attributes."""
        packet = _unpack_structure(payload, ARCH_RAY_HEADER)

        if self._bit_mask is None:
            # this is the first ray_hdr block read
            self.ngates = packet["gates"]
            self._bit_mask = packet["bit_mask"]
            self._field_nums = [b for b in range(38) if self._bit_mask & 2**b]
            self._dtype = ",".join(
                [DATA_FORMAT[self.field_info[i]["format"]] for i in self._field_nums]
            )
            self._ray_bsize = np.dtype(self._dtype).itemsize * packet["gates"]
        else:
            # check that the bit_mask and number of gates are constant
            if packet["bit_mask"] != self._bit_mask:
                raise NotImplementedError("bit_mask is not consistent.")
            if packet["gates"] != self.ngates:
                raise NotImplementedError("number of gates vary.")

        # store ray data and pointing data
        self._dstring += self._fh.read(self._ray_bsize)
        if self._include_ns_time:
            self.time.append(packet["time"] + packet["ns_time"] / 1e9)
        else:
            self.time.append(packet["time"])
        self.azimuth.append(packet["azimuth"])
        self.elevation.append(packet["elevation"])
        self._rays_in_current_sweep += 1

        return packet

    def _extract_fields(self):
        """Extract field data from _dstring attribute post read."""
        all_data = np.frombuffer(self._dstring, dtype=self._dtype)
        all_data = all_data.reshape(-1, self.ngates)
        for i, field_num in enumerate(self._field_nums):

            fdata = np.ma.masked_values(all_data[all_data.dtype.names[i]], 0)
            # apply scale and offset factors to interger data types
            if issubclass(fdata.dtype.type, np.integer):
                dat_factor = float(self.field_info[field_num]["dat_factor"])
                dat_bias = float(self.field_info[field_num]["dat_bias"])
                fld_factor = float(self.field_info[field_num]["fld_factor"])
                fdata = (fdata * dat_factor + dat_bias) / fld_factor

            self.fields[field_num] = fdata
        return


# CHL packet types
ARCH_FORMAT_VERSION = 0x00010000
ARCH_ID_CONTROL = 0x5AA80001
ARCH_ID_FIELD_SCALE = 0x5AA80002
ARCH_ID_RAY_HDR = 0x5AA80003
ARCH_ID_FILE_HDR = 0x5AA80004
ARCH_ID_SWEEP_BLOCK = 0x5AA80005
HSK_ID_PROCESSOR_INFO = 0x5AA50003
HSK_ID_RADAR_INFO = 0x5AA50001
HSK_ID_SCAN_SEG = 0x5AA50002

# Additional constants
SCAN_MODE_NAMES = ["ppi", "rhi", "fixed", "manual ppi", "manual rhi", "idle"]
DATA_FORMAT = ["uint8", "uint64", "float32", "uint16"]

##############
# Structures #
##############


def _unpack_structure(string, structure):
    """Unpack a structure."""
    fmt = "".join([i[1] for i in structure])
    tpl = struct.unpack(fmt, string)
    return dict(zip([i[0] for i in structure], tpl))


ARCH_FILE_HDR_T = (
    ("version", "I"),
    ("creation_version", "I"),
    ("creator_id", "32s"),
    ("sweep_table_offset", "Q"),
)

ARCH_RAY_HEADER = (
    ("azimuth", "f"),
    ("elevation", "f"),
    ("azimuth_width", "f"),
    ("elevation_width", "f"),
    ("gates", "H"),
    ("beam_index", "H"),
    ("ns_time", "I"),
    ("time", "Q"),
    ("bit_mask", "Q"),
    ("ray_number", "I"),
    ("num_pulses", "I"),
)

FIELD_SCALE_T = (
    ("format", "i"),
    ("min_val", "f"),
    ("max_val", "f"),
    ("bit_mask_pos", "i"),
    ("type_hint", "i"),
    ("fld_factor", "i"),
    ("dat_factor", "i"),
    ("dat_bias", "i"),
    ("name", "32s"),
    ("units", "32s"),
    ("descr", "128s"),
)

RADAR_INFO_T = (
    ("radar_name", "32s"),
    ("latitude", "f"),
    ("longitude", "f"),
    ("altitude", "f"),
    ("beamwidth", "f"),
    ("wavelength_cm", "f"),
    ("unused1", "f"),
    ("unused2", "f"),
    ("unused3", "f"),
    ("unused4", "f"),
    ("gain_ant_h", "f"),
    ("gain_ant_v", "f"),
    ("zdr_cal_base", "f"),
    ("phidp_rot", "f"),
    ("base_radar_constant", "f"),
    ("unused5", "f"),
    ("power_measurement_loss_h", "f"),
    ("power_measurement_loss_v", "f"),
    ("zdr_cal_base_vhs", "f"),
    ("test_power_h", "f"),
    ("test_power_v", "f"),
    ("dc_loss_h", "f"),
    ("dc_loss_v", "f"),
)

PROCESSOR_INFO = (
    ("polarization_mode", "i"),
    ("processing_mode", "i"),
    ("pulse_type", "i"),
    ("test_type", "i"),
    ("integration_cycle_pulses", "I"),
    ("clutter_filter_number", "I"),
    ("range_gate_averaging", "I"),
    ("indexed_beam_width", "f"),
    ("gate_spacing", "f"),
    ("prt_usec", "f"),
    ("range_start", "f"),
    ("range_stop", "f"),
    ("max_gate", "I"),
    ("test_power", "f"),
    ("unused1", "f"),
    ("unused2", "f"),
    ("test_pulse_range", "f"),
    ("test_pulse_length", "f"),
    ("prt2", "f"),
    ("range_offset", "f"),
)

SCAN_SEG = (
    ("az_manual", "f"),
    ("el_manual", "f"),
    ("az_start", "f"),
    ("el_start", "f"),
    ("scan_rate", "f"),
    ("segname", "24s"),
    ("opt", "i"),
    ("follow_mode", "i"),
    ("scan_type", "i"),
    ("scan_flags", "I"),
    ("volume_num", "I"),
    ("sweep_num", "I"),
    ("time_limit", "I"),
    ("webtilt", "I"),
    ("left_limit", "f"),
    ("right_limit", "f"),
    ("up_limit", "f"),
    ("down_limit", "f"),
    ("step", "f"),
    ("max_sweeps", "I"),
    ("filter_break_sweep", "I"),
    ("clutter_filter1", "I"),
    ("clutter_filter2", "I"),
    ("project", "16s"),
    ("current_fixed_angle", "f"),
)



"""
=================================
Plot a Cross Section from a Grid
=================================

This is an example of how to plot a cross section
of your radar grid using the GridMapDisplay

"""
print(__doc__)

# Author: Max Grover (mgrover@anl.gov)
# License: BSD 3 clause

import cartopy.crs as ccrs
import matplotlib.pyplot as plt

import pyart
from pyart.testing import get_test_data

# Read in the data from two XSAPR radars
xsapr_sw_file = get_test_data("swx_20120520_0641.nc")
xsapr_se_file = get_test_data("sex_20120520_0641.nc")
radar_sw = pyart.io.read_cfradial(xsapr_sw_file)
radar_se = pyart.io.read_cfradial(xsapr_se_file)

# Filter out gates with reflectivity > 100 from both radars
gatefilter_se = pyart.filters.GateFilter(radar_se)
gatefilter_se.exclude_transition()
gatefilter_se.exclude_above("corrected_reflectivity_horizontal", 100)
gatefilter_sw = pyart.filters.GateFilter(radar_sw)
gatefilter_sw.exclude_transition()
gatefilter_sw.exclude_above("corrected_reflectivity_horizontal", 100)

# perform Cartesian mapping, limit to the reflectivity field.
grid = pyart.map.grid_from_radars(
    (radar_se, radar_sw),
    gatefilters=(gatefilter_se, gatefilter_sw),
    grid_shape=(20, 181, 181),
    grid_limits=((500, 10000), (-50000, 40000), (-60000, 40000)),
    grid_origin=(36.57861, -97.363611),
    fields=["corrected_reflectivity_horizontal"],
)

# Define some start and end points, using (latitude, longitude)
start = (36.7, -97.7)
end = (36.2, -97.8)

# Setup the figure, and plot our x/y view of the radar
fig = plt.figure(figsize=(18, 6))
ax1 = plt.subplot(121, projection=ccrs.PlateCarree())
display = pyart.graph.GridMapDisplay(grid)
display.plot_grid(
    "corrected_reflectivity_horizontal",
    ax=ax1,
    cmap="pyart_HomeyerRainbow",
    vmin=-20,
    vmax=70,
)

# Plot our start and end points, as well as a line in between the two
ax1.scatter(start[1], start[0], color="tab:blue", label="Start")
ax1.scatter(end[1], end[0], color="black", label="End")
ax1.plot([start[1], end[1]], [start[0], end[0]], color="k", linestyle=":")
plt.legend(loc="upper right")

# Add a cross section, using our start and end points, and set our x-axis as latitude (lat)
ax2 = plt.subplot(122)
display.plot_cross_section(
    "corrected_reflectivity_horizontal",
    start,
    end,
    x_axis="lat",
    cmap="pyart_HomeyerRainbow",
    vmin=-20,
    vmax=70,
)



"""
Common graphing routines.

"""

import matplotlib.pyplot as plt
from netCDF4 import num2date

from ..config import get_field_colormap, get_field_limits

########################
# Common radar methods #
########################


def parse_ax(ax):
    """Parse and return ax parameter."""
    if ax is None:
        ax = plt.gca()
    return ax


def parse_ax_fig(ax, fig):
    """Parse and return ax and fig parameters."""
    if ax is None:
        ax = plt.gca()
    if fig is None:
        fig = plt.gcf()
    return ax, fig


def parse_cmap(cmap, field=None):
    """Parse and return the cmap parameter."""
    if cmap is None:
        cmap = get_field_colormap(field)
    return cmap


def parse_vmin_vmax(container, field, vmin, vmax):
    """Parse and return vmin and vmax parameters."""
    field_dict = container.fields[field]
    field_default_vmin, field_default_vmax = get_field_limits(field)
    if vmin is None:
        if "valid_min" in field_dict:
            vmin = field_dict["valid_min"]
        else:
            vmin = field_default_vmin
    if vmax is None:
        if "valid_max" in field_dict:
            vmax = field_dict["valid_max"]
        else:
            vmax = field_default_vmax
    return vmin, vmax


def parse_lon_lat(grid, lon, lat):
    """Parse lat and lon parameters"""
    if lat is None:
        lat = grid.origin_latitude["data"][0]
    if lon is None:
        lon = grid.origin_longitude["data"][0]
    return lon, lat


def generate_colorbar_label(standard_name, units):
    """Generate and return a label for a colorbar."""
    return str(standard_name).replace("_", " ") + " (" + units + ")"


def generate_field_name(container, field):
    """Return a nice field name for a particular field."""
    if "standard_name" in container.fields[field]:
        field_name = container.fields[field]["standard_name"]
    elif "long_name" in container.fields[field]:
        field_name = container.fields[field]["long_name"]
    else:
        field_name = str(field)
    field_name = field_name.replace("_", " ")
    field_name = field_name[0].upper() + field_name[1:]
    return field_name


def generate_radar_name(radar):
    """Return radar name."""
    if "instrument_name" in radar.metadata:
        return radar.metadata["instrument_name"]
    else:
        return ""


def generate_grid_name(grid):
    """Return grid name."""
    if "instrument_name" in grid.metadata:
        return grid.metadata["instrument_name"]
    else:
        return ""


def generate_radar_time_begin(radar):
    """Return time begin in datetime instance."""
    # datetime object describing first sweep time
    times = int(radar.time["data"][0])
    units = radar.time["units"]
    calendar = radar.time["calendar"]
    return num2date(
        times,
        units,
        calendar,
        only_use_cftime_datetimes=False,
        only_use_python_datetimes=True,
    )


def generate_radar_time_sweep(radar, sweep):
    """Return time that a specific sweep began in a datetime instance."""
    first_ray = radar.sweep_start_ray_index["data"][sweep]
    times = int(radar.time["data"][first_ray])
    units = radar.time["units"]
    calendar = radar.time["calendar"]
    return num2date(
        times,
        units,
        calendar,
        only_use_cftime_datetimes=False,
        only_use_python_datetimes=True,
    )


def generate_grid_time_begin(grid):
    """Return time begin in datetime instance."""
    times = int(grid.time["data"][0])
    units = grid.time["units"]
    if "calendar" in grid.time:
        calendar = grid.time["calendar"]
    else:
        calendar = "standard"
    return num2date(
        times,
        units,
        calendar,
        only_use_cftime_datetimes=False,
        only_use_python_datetimes=True,
    )


def generate_filename(
    radar, field, sweep, ext="png", datetime_format="%Y%m%d%H%M%S", use_sweep_time=False
):
    """
    Generate a filename for a plot.

    Generated filename has form:
        radar_name_field_sweep_time.ext

    Parameters
    ----------
    radar : Radar
        Radar structure.
    field : str
        Field plotted.
    sweep : int
        Sweep plotted.
    ext : str
        Filename extension.
    datetime_format : str
        Format of datetime (using strftime format).
    use_sweep_time : bool
        If true, the current sweep's beginning time is used.

    Returns
    -------
    filename : str
        Filename suitable for saving a plot.

    """
    name_s = generate_radar_name(radar).replace(" ", "_")
    field_s = field.replace(" ", "_")
    if use_sweep_time:
        time_s = generate_radar_time_sweep(radar, sweep).strftime(datetime_format)
    else:
        time_s = generate_radar_time_begin(radar).strftime(datetime_format)
    sweep_s = str(sweep).zfill(2)
    return f"{name_s}_{field_s}_{sweep_s}_{time_s}.{ext}"


def generate_grid_filename(grid, field, level, ext="png"):
    """
    Generate a filename for a plot.

    Generated filename has form:
        grid_name_field_level_time.ext

    Parameters
    ----------
    grid : Grid
        Grid structure.
    field : str
        Field plotted.
    level : int
        Level plotted.
    ext : str
        Filename extension.

    Returns
    -------
    filename : str
        Filename suitable for saving a plot.

    """
    name_s = generate_grid_name(grid).replace(" ", "_")
    field_s = field.replace(" ", "_")
    time_s = generate_grid_time_begin(grid).strftime("%Y%m%d%H%M%S")
    level_s = str(level).zfill(2)
    return f"{name_s}_{field_s}_{level_s}_{time_s}.{ext}"


def generate_cross_section_title(grid, field, start, end):
    """
    Generate a title for a plot.

    Parameters
    ----------
    grid : Grid
        Radar structure.
    field : str
        Field plotted.
    start : tuple
        Latitude-Longitude pair of starting points
    end: tuple
        Latitude-Longitude pair of ending poins

    Returns
    -------
    title : str
        Plot title.

    """

    # Grab information about the grid
    time_str = generate_grid_time_begin(grid).strftime("%Y-%m-%dT%H:%M:%SZ")
    field_name = generate_field_name(grid, field)

    # Deal with the latitude/longitude part
    start_lat, start_lon = start
    end_lat, end_lon = end

    degree_symbol = "\N{DEGREE SIGN}"

    # Add degrees latitude/longitude labels

    if start_lat < 0:
        start_lat_string = f"{-start_lat}{degree_symbol}S"
    else:
        start_lat_string = f"{start_lat}{degree_symbol}N"

    if end_lat < 0:
        end_lat_string = f"{-end_lat}{degree_symbol}S"
    else:
        end_lat_string = f"{end_lat}{degree_symbol}N"

    if start_lon < 0:
        start_lon_string = f"{-start_lon}{degree_symbol}W"
    else:
        start_lon_string = f"{start_lon}{degree_symbol}E"

    if end_lon < 0:
        end_lon_string = f"{-end_lon}{degree_symbol}W"
    else:
        end_lon_string = f"{end_lon}{degree_symbol}E"

    return (
        f"({start_lat_string}, {start_lon_string})"
        f" to ({end_lat_string}, {end_lon_string}) \n {time_str} \n {field_name}"
    )


def generate_title(radar, field, sweep, datetime_format=None, use_sweep_time=True):
    """
    Generate a title for a plot.

    Parameters
    ----------
    radar : Radar
        Radar structure.
    field : str
        Field plotted.
    sweep : int
        Sweep plotted.
    datetime_format : str
        Format of datetime (using strftime format).
    use_sweep_time : bool
        If true, the current sweep's beginning time is used.

    Returns
    -------
    title : str
        Plot title.

    """
    if use_sweep_time:
        begin_time = generate_radar_time_sweep(radar, sweep)
    else:
        begin_time = generate_radar_time_begin(radar)
    if datetime_format:
        time_str = begin_time.strftime(datetime_format)
    else:
        time_str = begin_time.isoformat() + "Z"
    fixed_angle = radar.fixed_angle["data"][sweep]
    l1 = f"{generate_radar_name(radar)} {fixed_angle:.1f} Deg. {time_str} "
    field_name = generate_field_name(radar, field)
    return l1 + "\n" + field_name


def generate_grid_title(grid, field, level):
    """
    Generate a title for a plot.

    Parameters
    ----------
    grid : Grid
        Radar structure.
    field : str
        Field plotted.
    level : int
        Verical level plotted.


    Returns
    -------
    title : str
        Plot title.

    """
    time_str = generate_grid_time_begin(grid).isoformat() + "Z"
    height = grid.z["data"][level] / 1000.0
    l1 = f"{generate_grid_name(grid)} {height:.1f} km {time_str} "
    field_name = generate_field_name(grid, field)
    return l1 + "\n" + field_name


def generate_longitudinal_level_title(grid, field, level):
    """
    Generate a title for a plot.

    Parameters
    ----------
    grid : Grid
        Radar structure.
    field : str
        Field plotted.
    level : int
        Longitudinal level plotted.


    Returns
    -------
    title : str
        Plot title.

    """
    time_str = generate_grid_time_begin(grid).strftime("%Y-%m-%dT%H:%M:%SZ")
    disp = grid.x["data"][level] / 1000.0
    if disp >= 0:
        direction = "east"
    else:
        direction = "west"
        disp = -disp
    l1 = "{} {:.1f} km {} of origin {} ".format(
        generate_grid_name(grid), disp, direction, time_str
    )
    field_name = generate_field_name(grid, field)
    return l1 + "\n" + field_name


def generate_latitudinal_level_title(grid, field, level):
    """
    Generate a title for a plot.

    Parameters
    ----------
    grid : Grid
        Radar structure.
    field : str
        Field plotted.
    level : int
        Latitudinal level plotted.


    Returns
    -------
    title : str
        Plot title.

    """
    time_str = generate_grid_time_begin(grid).isoformat() + "Z"
    disp = grid.y["data"][level] / 1000.0
    if disp >= 0:
        direction = "north"
    else:
        direction = "south"
        disp = -disp
    l1 = "{} {:.1f} km {} of origin {} ".format(
        generate_grid_name(grid), disp, direction, time_str
    )
    field_name = generate_field_name(grid, field)
    return l1 + "\n" + field_name


def generate_vpt_title(radar, field):
    """
    Generate a title for a VPT plot.

    Parameters
    ----------
    radar : Radar
        Radar structure.
    field : str
        Field plotted.

    Returns
    -------
    title : str
        Plot title.

    """
    time_str = generate_radar_time_begin(radar).isoformat() + "Z"
    l1 = f"{generate_radar_name(radar)} {time_str} "
    field_name = generate_field_name(radar, field)
    return l1 + "\n" + field_name


def generate_ray_title(radar, field, ray):
    """
    Generate a title for a ray plot.

    Parameters
    ----------
    radar : Radar
        Radar structure.
    field : str
        Field plotted.
    ray : int
        Ray plotted.

    Returns
    -------
    title : str
        Plot title.

    """
    time_str = generate_radar_time_begin(radar).isoformat() + "Z"
    l1 = f"{generate_radar_name(radar)} {time_str}"
    azim = radar.azimuth["data"][ray]
    elev = radar.elevation["data"][ray]
    l2 = "Ray: %i  Elevation: %.1f Azimuth: %.1f" % (ray, elev, azim)
    field_name = generate_field_name(radar, field)
    return l1 + "\n" + l2 + "\n" + field_name


def generate_az_rhi_title(radar, field, azimuth):
    """
    Generate a title for a pseudo-RHI from PPI azimuth plot.

    Parameters
    ----------
    radar : Radar
        Radar structure.
    field : str
        Field plotted.
    azimuth : float
        Azimuth plotted.

    Returns
    -------
    title : str
        Plot title.

    """
    time_str = generate_radar_time_begin(radar).isoformat() + "Z"
    l1 = f"{generate_radar_name(radar)} {time_str} "
    l2 = "Azimuth: %.1f deg" % azimuth
    field_name = generate_field_name(radar, field)
    return l1 + "\n" + l2 + "\n" + field_name


def set_limits(xlim=None, ylim=None, ax=None):
    """
    Set the display limits.

    Parameters
    ----------
    xlim : tuple, optional
        2-Tuple containing y-axis limits in km. None uses default limits.
    ylim : tuple, optional
        2-Tuple containing x-axis limits in km. None uses default limits.
    ax : Axis
        Axis to adjust.  None will adjust the current axis.

    """
    if ax is None:
        ax = plt.gca()
    if ylim is not None:
        ax.set_ylim(ylim)
    if xlim is not None:
        ax.set_xlim(xlim)



"""
==================================================================
Reading Older NEXRAD Data and Fixing Latitude and Longitude Issues
==================================================================

In this example, we will show how to read in older NEXRAD files prior
to 2008 that are missing some coordinate metadata.

"""
print(__doc__)


# Author: Zachary Sherman (zsherman@anl.gov)
# License: BSD 3 clause

######################################
# Import our required packages.

import cartopy.crs as ccrs
import matplotlib.pyplot as plt

import pyart

######################################
# Read older NEXRAD Level 2 Data
# ------------------------------
#
# Older NEXRAD files prior to 2008, have the tendency to not contain some of
# the required metadata for Py-ART's NEXRAD reader. This usually results in
# missing latitude and longitude data, so after reading with Py-ART, both
# coordinates have a value of 0. This example, we will show how to properly
# read in an older NEXRAD file.
#
# First we want to get an older file from amazon web service:
#
#    ``s3://noaa-nexrad-level2/year/month/date/radarsite/{radarsite}{year}{month}{date}_{hour}{minute}{second}.gz``
#
# Where in our case, we are using a sample data file from Handford, CA (KHNX)
# on July 24, 2006, at 0203:38 UTC. This means our path would look like this:

# Note: Older files do not contain the 'V06' but instead '.gz' in the AWS path.

aws_nexrad_level2_file = (
    "s3://noaa-nexrad-level2/2006/07/24/KHNX/KHNX20060724_020338.gz"
)

######################################
# We can use the **pyart.io.read_nexrad_archive** module to access our data, passing in the filepath.

radar = pyart.io.read_nexrad_archive(aws_nexrad_level2_file)

######################################
# Now let us take a look at the radar latitude and longitude data.
print(radar.latitude["data"])
print(radar.longitude["data"])

######################################
# This is clearly not correct! The problem is the reader could not find the
# metadata (message 31) for the coordinates.
#
# Lucky for us, we can provide the station in Py-ART's NEXRAD reader, which will
# pull the coordinate information from a dictionary found within Py-ART.

radar = pyart.io.read_nexrad_archive(aws_nexrad_level2_file, station="KHNX")

######################################
# Again, let us take a look at the radar latitude and longitude data.
print(radar.latitude["data"])
print(radar.longitude["data"])

##########################################
# Everything now looks correct as this is in Handford CA!

# We can create a plot as well utilizing Cartopy to see how it looks.
display = pyart.graph.RadarMapDisplay(radar)

# Setting projection and ploting the first tilt.
projection = ccrs.LambertConformal(
    central_latitude=radar.latitude["data"][0],
    central_longitude=radar.longitude["data"][0],
)

fig = plt.figure(figsize=(6, 6))
display.plot_ppi_map(
    "reflectivity", 0, vmin=-20, vmax=54, projection=projection, resolution="10m"
)



"""
=======================================
Plot a cross section from  a PPI volume
=======================================

An example which extracts a cross section at two azimuth angles from a volume
of PPI scans and plots both cross sections.

"""
print(__doc__)

# Author: Jonathan J. Helmus (jhelmus@anl.gov)
# License: BSD 3 clause

import matplotlib.pyplot as plt

import pyart
from pyart.testing import get_test_data

# Read the data, a cfradial file
filename = get_test_data("swx_20120520_0641.nc")
radar = pyart.io.read(filename)

# Create a cross section at 225 and 270 degrees azimuth
xsect = pyart.util.cross_section_ppi(radar, [225, 270])

# Set the colorbar label
colorbar_label = "Equivalent \n reflectivity factor \n (dBZ)"

display = pyart.graph.RadarDisplay(xsect)
fig = plt.figure()
ax1 = fig.add_subplot(211)
display.plot(
    "reflectivity_horizontal", 0, vmin=-32, vmax=64.0, colorbar_label=colorbar_label
)
plt.ylim(0, 15)
ax2 = fig.add_subplot(212)
display.plot(
    "reflectivity_horizontal", 1, vmin=-32, vmax=64.0, colorbar_label=colorbar_label
)
plt.ylim(0, 15)

plt.tight_layout()
plt.show()



""" Unit Tests for Py-ART's io/mdv_grid.py module. """

import datetime
from io import BytesIO

import numpy as np
from numpy.testing import assert_almost_equal, assert_raises, assert_warns

import pyart


class Mdv_grid_Tests:
    """
    Class for declaring unit tests for the io/mdv_grid.py module which
    require a temporary file which will be removed at the end of the test.
    """

    def test_write_read_target(self):
        # write and read in target grid
        original_grid = pyart.testing.make_target_grid()
        tmpfile = BytesIO()
        pyart.io.write_grid_mdv(tmpfile, original_grid)
        tmpfile.seek(0)
        grid = pyart.io.read_grid_mdv(tmpfile, file_field_names=True)

        # check the contents of the grid which was read
        self.check_target_reflectivity_field(grid)
        attrs_to_check = [
            "time",
            "origin_latitude",
            "origin_longitude",
            "origin_altitude",
            "x",
            "y",
            "z",
        ]
        for attr in attrs_to_check:
            self.check_attr_dics(grid, original_grid, attr)

    @staticmethod
    def check_target_reflectivity_field(grid):
        # check the reflectivity field data
        assert "reflectivity" in grid.fields
        fdata = grid.fields["reflectivity"]["data"]

        assert fdata.shape == (2, 400, 320)
        assert fdata.dtype == np.float32

        assert np.abs(fdata[0, 0, 0] - 0.0) <= 0.01
        assert np.abs(fdata[0, 60, 50] - 10.0) <= 0.01
        assert np.abs(fdata[0, 110, 90] - 20.0) <= 0.01
        assert np.abs(fdata[0, 160, 130] - 30.0) <= 0.01

        assert np.abs(fdata[1, 0, 0] - 5.0) <= 0.01
        assert np.abs(fdata[1, 60, 50] - 15.0) <= 0.01
        assert np.abs(fdata[1, 110, 90] - 25.0) <= 0.01
        assert np.abs(fdata[1, 160, 130] - 35.0) <= 0.01

    @staticmethod
    def check_attr_dics(grid1, grid2, attr_name):
        print("Checking:", attr_name)
        dic1 = getattr(grid1, attr_name)
        dic2 = getattr(grid2, attr_name)
        # check that the data and units keys are similar
        assert np.allclose(dic1["data"], dic2["data"])
        assert dic1["data"].shape == dic2["data"].shape
        assert dic1["data"].dtype == dic2["data"].dtype
        if "units" in dic1:
            assert dic1["units"] == dic2["units"]

    def test_write_read_target_delay(self):
        # write and read in the target grid with delayed field loading
        original_grid = pyart.testing.make_target_grid()
        tmpfile = BytesIO()
        pyart.io.write_grid_mdv(tmpfile, original_grid)
        tmpfile.seek(0)
        grid = pyart.io.read_grid_mdv(
            tmpfile, file_field_names=True, delay_field_loading=True
        )

        # check the contents of the grid which was read
        self.check_target_reflectivity_field(grid)
        attrs_to_check = [
            "time",
            "origin_latitude",
            "origin_longitude",
            "origin_altitude",
            "x",
            "y",
            "z",
        ]
        for attr in attrs_to_check:
            self.check_attr_dics(grid, original_grid, attr)

    def test_write_read_target_modified(self):
        # write and read in target grid with modifications
        original_grid = pyart.testing.make_target_grid()
        original_grid.metadata["radar_0_lon"] = -98.1
        original_grid.metadata["radar_0_lat"] = 36.74
        original_grid.metadata["radar_0_alt"] = 300.0
        original_grid.metadata["instrument_name"] = "testtesttest"
        original_grid.fields["reflectivity"]["standard_name"] = "fieldname"
        tmpfile = BytesIO()
        pyart.io.write_grid_mdv(tmpfile, original_grid)
        tmpfile.seek(0)
        grid = pyart.io.read_grid_mdv(tmpfile, file_field_names=True)

        # check the contents of the grid which was read
        self.check_target_reflectivity_field(grid)
        attrs_to_check = [
            "time",
            "origin_latitude",
            "origin_longitude",
            "origin_altitude",
            "x",
            "y",
            "z",
        ]
        for attr in attrs_to_check:
            self.check_attr_dics(grid, original_grid, attr)
        assert grid.metadata["instrument_name"] == "testtesttest"

    def test_read_write_mask(self):
        # Read and write a grid with field which is partially masked
        original_grid = pyart.testing.make_empty_grid(
            (20, 20, 20), ((0, 1), (0, 1), (0, 1))
        )
        fdata = np.ma.ones((20, 20, 20), dtype=np.float32)
        fdata[0, 0, 1] = np.ma.masked
        original_grid.fields["field_one"] = {"data": fdata}
        tmpfile = BytesIO()
        pyart.io.write_grid_mdv(tmpfile, original_grid)
        tmpfile.seek(0)

        # check the new grid object
        grid = pyart.io.read_grid_mdv(tmpfile, file_field_names=True)
        assert "field_one" in grid.fields
        fdata = grid.fields["field_one"]["data"]
        assert fdata.dtype == np.float32
        assert fdata.shape == (20, 20, 20)
        assert np.ma.isMA(fdata)
        assert np.ma.is_masked(fdata[0, 0, 1])
        assert not np.ma.is_masked(fdata[0, 0, 0])

    def test_raise_warning_to_many_z(self):
        # Grids greater than 122 level should raise a warning and
        # write out the first 122 levels.
        grid = pyart.testing.make_empty_grid((123, 2, 2), ((0, 1), (0, 1), (0, 1)))
        grid.fields["reflectivity"] = {"data": np.zeros((123, 2, 2))}
        tmpfile = BytesIO()
        assert_warns(UserWarning, pyart.io.write_grid_mdv, tmpfile, grid)
        tmpfile.seek(0)
        rgrid = pyart.io.read_grid_mdv(tmpfile)
        assert len(grid.z["data"]) == 123
        assert len(rgrid.z["data"]) == 122

    def test_write_types(self):
        # Write various data types
        original_grid = pyart.testing.make_empty_grid(
            (2, 2, 2), ((0, 1), (0, 1), (0, 1))
        )

        uint8_data = np.ones((2, 2, 2), dtype=np.float32) * 100.0 + 5
        original_grid.fields["field_one"] = {
            "data": uint8_data,
            "scale_factor": 10.0,
            "add_offset": 5.0,
            "_Write_as_dtype": "uint8",
            "_FillValue": 128,
        }

        uint16_data = np.ones((2, 2, 2), dtype=np.float32) * 22.0
        original_grid.fields["field_two"] = {
            "data": uint16_data,
            "_Write_as_dtype": "uint16",
            "missing_value": 256,
        }

        float32_data = np.ones((2, 2, 2), dtype=np.float32) * 98.0 + 2.0
        original_grid.fields["field_three"] = {
            "data": float32_data,
            "scale_factor": 10.0,
            "add_offset": 2.0,
        }
        tmpfile = BytesIO()
        pyart.io.write_grid_mdv(tmpfile, original_grid)
        tmpfile.seek(0)

        # verify that the data get read in correctly
        # use delay_field_loading to prevent closing of tmpfile
        grid = pyart.io.read_grid_mdv(
            tmpfile, file_field_names=True, delay_field_loading=True
        )
        assert np.all(grid.fields["field_one"]["data"] == 105)
        assert np.all(grid.fields["field_two"]["data"] == 22)
        assert np.all(grid.fields["field_three"]["data"] == 100)
        tmpfile.seek(0)

        # vertify that the data is encoded to the correct type in the file
        mdvfile = pyart.io.mdv_common.MdvFile(tmpfile)
        field_name = [h["field_name"] for h in mdvfile.field_headers]
        field_one = field_name.index("field_one")
        field_two = field_name.index("field_two")
        field_three = field_name.index("field_three")
        encoding = [h["encoding_type"] for h in mdvfile.field_headers]
        assert encoding[field_one] == pyart.io.mdv_common.ENCODING_INT8
        assert encoding[field_two] == pyart.io.mdv_common.ENCODING_INT16
        assert encoding[field_three] == pyart.io.mdv_common.ENCODING_FLOAT32
        mdvfile.close()

    def test_raise_error_on_invalid_dtype(self):
        #  Unsupported dtyped field should raise a TypeError
        grid = pyart.testing.make_empty_grid((2, 2, 2), ((0, 1), (0, 1), (0, 1)))
        grid.fields["reflectivity"] = {
            "data": np.zeros((2, 2, 2), dtype=np.float32),
            "_Write_as_dtype": "float16",
        }
        tmpfile = BytesIO()
        assert_raises(TypeError, pyart.io.write_grid_mdv, tmpfile, grid)


def test_time_dic_to_datetime():
    dt = pyart.io.mdv_grid._time_dic_to_datetime(
        {"data": [20], "units": "seconds since 1970-01-01 00:00:00"}
    )
    assert dt == datetime.datetime(1970, 1, 1, 0, 0, 20)

    dt = pyart.io.mdv_grid._time_dic_to_datetime(
        {
            "data": [20],
            "units": "seconds since 1970-01-01 00:00:00",
            "calendar": "standard",
        }
    )
    assert dt == datetime.datetime(1970, 1, 1, 0, 0, 20)


def test_mdv_degree_grid():
    grid = pyart.io.read_grid_mdv(pyart.testing.MDV_GRID_FILE, file_field_names=True)

    assert "refl" in grid.fields.keys()
    fdata = grid.fields["refl"]["data"]
    assert fdata.shape == (1, 1837, 3661)
    assert np.ma.is_masked(fdata[0, 0, 0])
    assert_almost_equal(fdata[0, 130, 2536], 20.0, 1)

    assert grid.x["units"] == "degree_E"
    assert_almost_equal(grid.x["data"][0], -129.99, 2)
    assert grid.y["units"] == "degree_N"
    assert_almost_equal(grid.y["data"][0], 20.01, 2)



""" Unit Tests for Py-ART's config.py module. """

import os
import warnings
from importlib import reload

import pyart

dirname = os.path.dirname(__file__)
CUSTOM_CONFIG_FILE = os.path.join(dirname, "custom_config.py")


def test_config_functions():
    pyart.load_config(CUSTOM_CONFIG_FILE)

    assert pyart.config.get_field_name("reflectivity") == "REF"
    assert pyart.config.get_fillvalue() == -5555.0

    metadata = pyart.config.get_metadata("azimuth")
    assert isinstance(metadata, dict)
    assert metadata["units"] == "foo"

    assert pyart.config.get_metadata("foobar") == {}

    pyart.load_config()  # load default
    assert pyart.config.get_field_name("reflectivity") == "reflectivity"


def test_filemetadata_custom():

    pyart.load_config(CUSTOM_CONFIG_FILE)

    # The MDV reflectivity field is mapped to velocity
    filemetadata = pyart.config.FileMetadata("mdv")
    assert filemetadata.get_field_name("DBZ_F") == "velocity"

    # there is no foobar field
    assert filemetadata.get_field_name("foobar") is None

    # and has a custom time metadata
    time = filemetadata("time")
    assert isinstance(time, dict)
    assert "foo" in time
    assert time["foo"] == "bar"

    # and standard elevation metadata
    elev = filemetadata("elevation")
    assert isinstance(elev, dict)
    assert "units" in elev
    assert elev["units"] == "degrees"

    assert filemetadata("foobar") == {}

    # invalid filetype does no field name mapping
    filemetadata = pyart.config.FileMetadata("foo")
    assert filemetadata.get_field_name("foobar") == "foobar"

    # additional metadata
    filemetadata = pyart.config.FileMetadata(
        "sigmet", additional_metadata={"baz": {"units": "baz_unit"}}
    )
    assert filemetadata("baz")["units"] == "baz_unit"

    # file_file_names
    filemetadata = pyart.config.FileMetadata("sigmet", file_field_names=True)
    assert filemetadata.get_field_name("DBT") == "DBT"

    # exclude fields
    filemetadata = pyart.config.FileMetadata(
        "sigmet", exclude_fields=["spectrum_width"]
    )
    assert filemetadata.get_field_name("WIDTH2") is None


def test_init_load():

    # load the custom config and verify
    os.environ["PYART_CONFIG"] = CUSTOM_CONFIG_FILE
    reload(pyart.config)
    assert pyart.config.get_field_name("reflectivity") == "REF"

    # starting up with an invalid config file raises a warning and
    # loads the default config
    os.environ["PYART_CONFIG"] = "nullnullnull"
    with warnings.catch_warnings(record=True) as w:
        reload(pyart.config)
        assert len(w) > 0
    assert pyart.config.get_field_name("reflectivity") == "reflectivity"

    # load the custom config and verify
    os.environ["PYART_CONFIG"] = CUSTOM_CONFIG_FILE
    reload(pyart.config)
    assert pyart.config.get_field_name("reflectivity") == "REF"

    # no PYART_CONFIG loads the default
    os.environ.pop("PYART_CONFIG")
    reload(pyart.config)
    assert pyart.config.get_field_name("reflectivity") == "reflectivity"


def test_intergration():

    pyart.load_config()  # load default
    radar = pyart.io.read_mdv(pyart.testing.MDV_PPI_FILE)
    assert "reflectivity" in radar.fields
    assert "velocity" not in radar.fields

    # MDV custom maps reflectivity to velocity, and has custom time metadata
    pyart.load_config(CUSTOM_CONFIG_FILE)
    radar = pyart.io.read_mdv(pyart.testing.MDV_PPI_FILE)
    assert "reflectivity" not in radar.fields
    assert "velocity" in radar.fields
    assert radar.time["foo"] == "bar"
    pyart.load_config()



"""
====================================================
Create a plot of multiple moments from a NEXRAD file
====================================================

An example which creates a plot containing multiple moments taken from a
NEXRAD Archive file.

"""
print(__doc__)

# Author: Jonathan J. Helmus (jhelmus@anl.gov)
# License: BSD 3 clause

import matplotlib.pyplot as plt

import pyart
from pyart.testing import get_test_data

filename = get_test_data("KATX20130717_195021_V06")
radar = pyart.io.read_nexrad_archive(filename)
display = pyart.graph.RadarDisplay(radar)
fig = plt.figure(figsize=(10, 10))

ax = fig.add_subplot(221)
display.plot(
    "velocity",
    1,
    ax=ax,
    title="Doppler Velocity",
    colorbar_label="",
    vmin=-32.0,
    vmax=32.0,
    axislabels=("", "North South distance from radar (km)"),
)
display.set_limits((-300, 300), (-300, 300), ax=ax)

ax = fig.add_subplot(222)
display.plot(
    "differential_reflectivity",
    0,
    ax=ax,
    title="Differential Reflectivity",
    colorbar_label="",
    axislabels=("", ""),
)
display.set_limits((-300, 300), (-300, 300), ax=ax)

ax = fig.add_subplot(223)
display.plot(
    "differential_phase", 0, ax=ax, title="Differential Phase", colorbar_label=""
)
display.set_limits((-300, 300), (-300, 300), ax=ax)

ax = fig.add_subplot(224)
display.plot(
    "cross_correlation_ratio",
    0,
    ax=ax,
    title="Correlation Coefficient",
    colorbar_label="",
    axislabels=("East West distance from radar (km)", ""),
)
display.set_limits((-300, 300), (-300, 300), ax=ax)

plt.show()



""" Unit Tests for Py-ART's io/nexrad_level3.py module. """

import numpy as np
from numpy.ma.core import MaskedArray

import pyart


def test_nexrad_level3_msg19():
    radar = pyart.io.read_nexrad_level3(pyart.testing.NEXRAD_LEVEL3_MSG19)

    assert radar.time["units"] == "seconds since 2015-01-02T02:05:28Z"
    assert radar.time["data"].shape == (360,)
    assert round(radar.time["data"][0]) == 0.0

    assert radar.range["data"].shape == (230,)
    assert round(radar.range["data"][100]) == 99900

    assert radar.scan_type == "ppi"

    assert radar.latitude["data"].shape == (1,)
    assert round(radar.latitude["data"][0]) == 33.0

    assert radar.longitude["data"].shape == (1,)
    assert round(radar.longitude["data"][0]) == -87.0

    assert radar.altitude["data"].shape == (1,)
    assert round(radar.altitude["data"][0]) == 231.0

    assert radar.altitude_agl is None

    assert radar.sweep_number["data"].shape == (1,)
    assert radar.sweep_number["data"][0] == 0

    assert radar.sweep_mode["data"].shape == (1,)
    assert radar.sweep_mode["data"].dtype.char == "S"
    assert np.all(radar.sweep_mode["data"] == [b"azimuth_surveillance"])

    assert radar.fixed_angle["data"].shape == (1,)
    assert round(radar.fixed_angle["data"][0], 2) == 0.50

    assert radar.sweep_start_ray_index["data"].shape == (1,)
    assert round(radar.sweep_start_ray_index["data"][0]) == 0.0

    assert radar.sweep_end_ray_index["data"].shape == (1,)
    assert round(radar.sweep_end_ray_index["data"][0]) == 359.0

    assert radar.target_scan_rate is None

    assert round(radar.azimuth["data"][0]) == 320.0
    assert round(radar.azimuth["data"][10]) == 330.0

    assert radar.elevation["data"].shape == (360,)
    assert round(radar.elevation["data"][0], 2) == 0.50

    assert radar.scan_rate is None
    assert radar.antenna_transition is None
    assert radar.instrument_parameters is None
    assert radar.radar_calibration is None

    assert radar.ngates == 230
    assert radar.nrays == 360
    assert radar.nsweeps == 1

    assert "reflectivity" in radar.fields.keys()
    assert radar.fields["reflectivity"]["data"].shape == (360, 230)
    assert type(radar.fields["reflectivity"]["data"]) is MaskedArray
    assert round(radar.fields["reflectivity"]["data"][10, 10]) == np.float32(25)


def test_nexrad_level3_msg161():
    radar = pyart.io.read_nexrad_level3(pyart.testing.NEXRAD_LEVEL3_MSG163)

    assert radar.time["units"] == "seconds since 2015-01-02T02:05:28Z"
    assert radar.time["data"].shape == (360,)
    assert round(radar.time["data"][0]) == 0.0

    assert radar.range["data"].shape == (1200,)
    assert round(radar.range["data"][100]) == 24975.0

    assert radar.scan_type == "ppi"

    assert radar.latitude["data"].shape == (1,)
    assert round(radar.latitude["data"][0]) == 33.0

    assert radar.longitude["data"].shape == (1,)
    assert round(radar.longitude["data"][0]) == -87.0

    assert radar.altitude["data"].shape == (1,)
    assert round(radar.altitude["data"][0]) == 231.0

    assert radar.altitude_agl is None

    assert radar.sweep_number["data"].shape == (1,)
    assert radar.sweep_number["data"][0] == 0

    assert radar.sweep_mode["data"].shape == (1,)
    assert np.all(radar.sweep_mode["data"] == [b"azimuth_surveillance"])

    assert radar.fixed_angle["data"].shape == (1,)
    assert round(radar.fixed_angle["data"][0], 2) == 0.50

    assert radar.sweep_start_ray_index["data"].shape == (1,)
    assert round(radar.sweep_start_ray_index["data"][0]) == 0.0

    assert radar.sweep_end_ray_index["data"].shape == (1,)
    assert round(radar.sweep_end_ray_index["data"][0]) == 359.0

    assert radar.target_scan_rate is None

    assert round(radar.azimuth["data"][0]) == 329.0
    assert round(radar.azimuth["data"][10]) == 339.0

    assert radar.elevation["data"].shape == (360,)
    assert round(radar.elevation["data"][0], 2) == 0.50

    assert radar.scan_rate is None
    assert radar.antenna_transition is None
    assert radar.instrument_parameters is None
    assert radar.radar_calibration is None

    assert radar.ngates == 1200
    assert radar.nrays == 360
    assert radar.nsweeps == 1

    field_name = "specific_differential_phase"
    assert field_name in radar.fields.keys()
    assert radar.fields[field_name]["data"].shape == (360, 1200)
    assert type(radar.fields[field_name]["data"]) is MaskedArray
    assert round(radar.fields[field_name]["data"][103, 170]) == np.float32(2)


def test_nexrad_level3_msg161_fileobj():
    fh = open(pyart.testing.NEXRAD_LEVEL3_MSG163, "rb")
    radar = pyart.io.read_nexrad_level3(fh)
    fh.close()
    field_name = "specific_differential_phase"
    assert field_name in radar.fields.keys()
    assert radar.fields[field_name]["data"].shape == (360, 1200)
    assert type(radar.fields[field_name]["data"]) is MaskedArray
    assert round(radar.fields[field_name]["data"][103, 170]) == np.float32(2)


def test_nexrad_level3_msg176():
    radar = pyart.io.read_nexrad_level3(pyart.testing.NEXRAD_LEVEL3_MSG176)

    assert radar.time["units"] == "seconds since 2020-03-19T18:01:43Z"
    assert radar.time["data"].shape == (360,)
    assert round(radar.time["data"][0]) == 0.0

    assert radar.range["data"].shape == (920,)
    assert round(radar.range["data"][100]) == 25125.0

    assert radar.scan_type == "ppi"

    assert radar.latitude["data"].shape == (1,)
    assert round(radar.latitude["data"][0]) == 42.0

    assert radar.longitude["data"].shape == (1,)
    assert round(radar.longitude["data"][0]) == -88.0

    assert radar.altitude["data"].shape == (1,)
    assert round(radar.altitude["data"][0]) == 232.0

    assert radar.altitude_agl is None

    assert radar.sweep_number["data"].shape == (1,)
    assert radar.sweep_number["data"][0] == 0

    assert radar.sweep_mode["data"].shape == (1,)
    assert np.all(radar.sweep_mode["data"] == [b"azimuth_surveillance"])

    assert radar.sweep_start_ray_index["data"].shape == (1,)
    assert round(radar.sweep_start_ray_index["data"][0]) == 0.0

    assert radar.sweep_end_ray_index["data"].shape == (1,)
    assert round(radar.sweep_end_ray_index["data"][0]) == 359.0

    assert radar.target_scan_rate is None

    assert round(radar.azimuth["data"][0]) == 0.0
    assert round(radar.azimuth["data"][10]) == 10.0

    assert radar.scan_rate is None
    assert radar.antenna_transition is None
    assert radar.instrument_parameters is None
    assert radar.radar_calibration is None

    assert radar.ngates == 920
    assert radar.nrays == 360
    assert radar.nsweeps == 1

    field_name = "radar_estimated_rain_rate"
    assert field_name in radar.fields.keys()
    assert radar.fields[field_name]["data"].shape == (360, 920)
    assert type(radar.fields[field_name]["data"]) is MaskedArray
    assert round(radar.fields[field_name]["data"][14, 14], 3) == np.float32(0.244)



"""
Routines for Ridgeline Instruments RXM-25 formatted NetCDF files.

"""

import datetime

try:
    import pytz

    _PYTZ_AVAILABLE = True
except ImportError:
    _PYTZ_AVAILABLE = False

import netCDF4
import numpy as np

import pyart

from ..config import get_metadata
from ..exceptions import MissingOptionalDependency
from ..testing import make_empty_ppi_radar


def read_rxm25(filename, cfradial_outfile=None, heading=None):
    """
    Read in Ridgeline Instruments RXM-25 formatted NetCDF data.

    Parameters
    ----------
    filename : str
        Name of Ridgeline Instruments (RLI) RXM-25 formatted NetCDF file
        from which to read data.
    cfradial_outfile : str, optional
        If file is to be converted to CF-Radial format, specify the output
        filename here.
    heading : float, optional
        If a heading offset exists, enter it here (in degrees).

    Returns
    -------
    radar : Radar
        Radar object.

    """
    if not _PYTZ_AVAILABLE:
        raise MissingOptionalDependency(
            "Pytz is required to use read_rxm25 but is " + "not installed"
        )

    data = netCDF4.Dataset(filename, "r")

    ngates = data.dimensions["Gate"].size
    rays_per_sweep = data.dimensions["Radial"].size
    radar = make_empty_ppi_radar(ngates, rays_per_sweep, 1)

    # Time needs to be converted from nss1970 to nss1989 and added to
    # Radar object.
    nineteen89 = datetime.datetime(1989, 1, 1, 0, 0, 1, tzinfo=pytz.utc)
    baseTime = np.array(
        [
            datetime.datetime.fromtimestamp(t, tz=pytz.UTC)
            for t in data.variables["Time"][:]
        ]
    )
    radar.time["data"] = np.array([t.total_seconds() for t in baseTime - nineteen89])

    if heading is not None:
        radar.heading = heading
        radar.azimuth["data"] = np.mod(data["Azimuth"][:] - radar.heading, 360.0)
    else:
        radar.azimuth["data"] = data["Azimuth"][:]

    radar.longitude["data"] = np.array([data.Longitude], dtype="float64")
    radar.latitude["data"] = np.array([data.Latitude], dtype="float64")
    radar.elevation["data"] = data["Elevation"][:]
    radar.altitude["data"] = np.array([data.Height], dtype="float64")

    fixed_agl_data = np.empty((1,), dtype="float32")
    fixed_agl_data[:] = np.mean(radar.elevation["data"][:rays_per_sweep])

    radar.fixed_angle["data"] = fixed_agl_data

    radar.range["data"] = np.linspace(
        data["StartRange"][0] / 1000,
        (ngates - 1) * data["GateWidth"][0] / 1000 + data["StartRange"][0] / 1000,
        ngates,
    )

    ref = data["Reflectivity"][:]
    norm_pow = data["NormalizedCoherentPower"][:]
    spec_w = data["SpectralWidth"][:]
    vel = data["Velocity"][:]
    corr_ref = data["CorrectedReflectivity"][:]
    diff_ref = data["DifferentialReflectivity"][:]
    diff_phase = data["DifferentialPhase"][:]
    spec_phase = data["SpecificPhase"][:]
    corr_diff_ref = data["CorrectedDifferentialReflectivity"][:]
    sig_noise = data["SignalToNoiseRatio"][:]
    rain_rate = data["RainfallRate"][:]
    cross_ra = data["CrossPolCorrelation"][:]

    fields = {
        "reflectivity": get_metadata("reflectivity"),
        "normalized_coherent_power": get_metadata("normalized_coherent_power"),
        "spectral_width": get_metadata("spectral_width"),
        "velocity": get_metadata("velocity"),
        "corrected_reflectivity": get_metadata("correct_reflectivity"),
        "differential_reflectivity": get_metadata("differential_reflectivity"),
        "differential_phase": get_metadata("differential_phase"),
        "specific_differential_phase": get_metadata("specific_differential_phase"),
        "corrected_differential_reflectivity": get_metadata(
            "corrected_differential_reflectivity"
        ),
        "signal_to_noise_ratio": get_metadata("signal_to_noise_ratio"),
        "rain_rate": get_metadata("rain_rate"),
        "cross_correlation_ratio": get_metadata("cross_correlation_ratio"),
    }

    radar.fields = fields
    radar.fields["reflectivity"]["data"] = ref
    radar.fields["normalized_coherent_power"]["data"] = norm_pow
    radar.fields["spectral_width"]["data"] = spec_w
    radar.fields["velocity"]["data"] = vel
    radar.fields["corrected_reflectivity"]["data"] = corr_ref
    radar.fields["differential_reflectivity"]["data"] = diff_ref
    radar.fields["differential_phase"]["data"] = diff_phase
    radar.fields["specific_differential_phase"]["data"] = spec_phase
    radar.fields["corrected_differential_reflectivity"]["data"] = corr_diff_ref
    radar.fields["signal_to_noise_ratio"]["data"] = sig_noise
    radar.fields["rain_rate"]["data"] = rain_rate
    radar.fields["cross_correlation_ratio"]["data"] = cross_ra

    radar.metadata["instrument_name"] = "RXM-25"

    if cfradial_outfile is not None:
        pyart.io.write_cfradial(cfradial_outfile, radar, arm_time_variables=True)

    return radar



# /usr/bin/env python
# dummify (replace all data with a single value) a NEXRAD Level II file.


import numpy as np

import pyart.io.nexrad_level2 as nexrad

NEXRAD_FILE = "KATX20130717_195021_V06"
OUTPUT_FILE = "KATX20130717_195021_V06_DUMMY"

# sizes of structures
RECORD_SIZE = 2432
MSG_HEADER_SIZE = 16
DATA_BLOCK_SIZE = 28

fin = open(NEXRAD_FILE, "rb")
out = open(OUTPUT_FILE, "wb")

# read in the first 134 records as write as is
out.write(fin.read(24 + 12 + RECORD_SIZE * 134))

# read the rest of the file, and create a array of int8
buf = fin.read()
buf_length = len(buf)
buf_data = np.frombuffer(buf, dtype="i1")
fin.close()

# replace all radial data with a single value so it compresses well
pos = 0
while pos < buf_length:

    # retrieve the record (msg31)
    new_pos, record = nexrad._get_record_from_buf(buf, pos)
    if record["header"]["type"] != 31:
        pos += RECORD_SIZE
        continue

    # dummy out all moments
    for bp_number in [4, 5, 6, 7, 8, 9]:
        bp_str = "block_pointer_" + str(bp_number)
        block_pointer = record["msg_header"][bp_str]
        if block_pointer != 0:
            bpos = pos + block_pointer + MSG_HEADER_SIZE
            d = nexrad._unpack_from_buf(buf, bpos, nexrad.GENERIC_DATA_BLOCK)
            points = d["ngates"]
            if d["data_name"] == "PHI":
                points *= 2
            start = pos + MSG_HEADER_SIZE + DATA_BLOCK_SIZE + block_pointer
            buf_data[start : start + points] = 2

    pos = new_pos

# write out the modified data
out.write(buf_data.tostring())
out.close()



"""
Routines for reading ARM vertically-pointing radar ingest (e.g., a1) files.
These files are characterized by being NetCDF files that do not fully conform
to the CF/Radial convention. Nonetheless this module borrows heavily from the
existing CF/Radial module.

"""

import netCDF4
import numpy as np

from ..config import FileMetadata
from ..core.radar import Radar
from ..io import cfradial


def read_kazr(
    filename,
    field_names=None,
    additional_metadata=None,
    file_field_names=False,
    exclude_fields=None,
    include_fields=None,
):
    """
    Read K-band ARM Zenith Radar (KAZR) NetCDF ingest data.

    Parameters
    ----------
    filename : str
        Name of NetCDF file to read data from.
    field_names : dict, optional
        Dictionary mapping field names in the file names to radar field names.
        Unlike other read functions, fields not in this dictionary or having a
        value of None are still included in the radar.fields dictionary, to
        exclude them use the `exclude_fields` parameter. Fields which are
        mapped by this dictionary will be renamed from key to value.
    additional_metadata : dict of dicts, optional
        This parameter is not used, it is included for uniformity.
    file_field_names : bool, optional
        True to force the use of the field names from the file in which
        case the `field_names` parameter is ignored. False will use to
        `field_names` parameter to rename fields.
    exclude_fields : list or None, optional
        List of fields to exclude from the radar object. This is applied
        after the `file_field_names` and `field_names` parameters. Set
        to None to include all fields specified by include_fields.
    include_fields : list or None, optional
        List of fields to include from the radar object. This is applied
        after the `file_field_names` and `field_names` parameters. Set
        to None to include all fields not specified by exclude_fields.

    Returns
    -------
    radar : Radar
        Radar object.
    """

    # create metadata retrieval object
    filemetadata = FileMetadata(
        "cfradial",
        field_names,
        additional_metadata,
        file_field_names,
        exclude_fields,
        include_fields,
    )

    # read the data
    ncobj = netCDF4.Dataset(filename)
    ncvars = ncobj.variables

    # 4.1 Global attribute -> move to metadata dictionary
    metadata = {k: getattr(ncobj, k) for k in ncobj.ncattrs()}
    metadata["n_gates_vary"] = "false"

    # 4.2 Dimensions (do nothing)

    # 4.3 Global variable -> move to metadata dictionary
    if "volume_number" in ncvars:
        metadata["volume_number"] = int(ncvars["volume_number"][:])
    else:
        metadata["volume_number"] = 0

    global_vars = {
        "platform_type": "fixed",
        "instrument_type": "radar",
        "primary_axis": "axis_z",
    }
    # ignore time_* global variables, these are calculated from the time
    # variable when the file is written.
    for var, default_value in global_vars.items():
        if var in ncvars:
            metadata[var] = str(netCDF4.chartostring(ncvars[var][:]))
        else:
            metadata[var] = default_value

    # 4.4 coordinate variables -> create attribute dictionaries
    time = cfradial._ncvar_to_dict(ncvars["time"])
    _range = cfradial._ncvar_to_dict(ncvars["range"])

    # 4.5 Ray dimension variables

    # 4.6 Location variables -> create attribute dictionaries
    # the only difference in this section to cfradial.read_cfradial is the
    # minor variable name differences:
    # latitude -> lat
    # longitude -> lon
    # altitdue -> alt
    latitude = cfradial._ncvar_to_dict(ncvars["lat"])
    longitude = cfradial._ncvar_to_dict(ncvars["lon"])
    altitude = cfradial._ncvar_to_dict(ncvars["alt"])

    # 4.7 Sweep variables -> create atrribute dictionaries
    # this is the section that needed the most work since the initial NetCDF
    # file did not contain any sweep information
    sweep_number = filemetadata("sweep_number")
    sweep_number["data"] = np.array([0], dtype=np.int32)

    sweep_mode = filemetadata("sweep_mode")
    sweep_mode["data"] = np.array(["vertical_pointing"], dtype=np.str)

    fixed_angle = filemetadata("fixed_angle")
    fixed_angle["data"] = np.array([90.0], dtype=np.float32)

    sweep_start_ray_index = filemetadata("sweep_start_ray_index")
    sweep_start_ray_index["data"] = np.array([0], dtype=np.int32)

    sweep_end_ray_index = filemetadata("sweep_end_ray_index")
    sweep_end_ray_index["data"] = np.array([ncvars["time"].size - 1], dtype=np.int32)

    # first sweep mode determines scan_type
    # this module is specific to vertically-pointing data
    scan_type = "vpt"

    # 4.8 Sensor pointing variables -> create attribute dictionaries
    # this section also required some changes since the initial NetCDF did not
    # contain any sensor pointing variables
    azimuth = filemetadata("azimuth")
    azimuth["data"] = 0.0 * np.ones(ncvars["time"].size, dtype=np.float32)

    elevation = filemetadata("elevation")
    elevation["data"] = 90.0 * np.ones(ncvars["time"].size, dtype=np.float32)

    # 4.9 Moving platform geo-reference variables

    # 4.10 Moments field data variables -> field attribute dictionary
    # all variables with dimensions of 'time', 'range' are fields
    keys = [k for k, v in ncvars.items() if v.dimensions == ("time", "range")]

    fields = {}
    for key in keys:
        field_name = filemetadata.get_field_name(key)
        if field_name is None:
            if exclude_fields is not None and key in exclude_fields:
                continue
            if include_fields is not None and key not in include_fields:
                continue
            field_name = key
        fields[field_name] = cfradial._ncvar_to_dict(ncvars[key])

    # 4.5 instrument_parameters sub-convention -> instrument_parameters dict
    # this section needed multiple changes and/or additions since the
    # instrument parameters were primarily located in the global attributes
    # this section is likely still incomplete
    omega = float(ncobj.radar_operating_frequency.split()[0])
    frequency = filemetadata("frequency")
    frequency["data"] = np.array([omega / 1e9], dtype=np.float32)

    prt_mode = filemetadata("prt_mode")
    prt_mode["data"] = np.array(["fixed"], dtype=np.str)

    prf = float(ncobj.pulse_repetition_frequency.split()[0])
    prt = filemetadata("prt")
    prt["data"] = (1.0 / prf) * np.ones(ncvars["time"].size, dtype=np.float32)

    v_nq = float(ncobj.nyquist_velocity.split()[0])
    nyquist_velocity = filemetadata("nyquist_velocity")
    nyquist_velocity["data"] = (v_nq * np.ones(ncvars["time"].size, dtype=np.float32),)
    samples = int(ncobj.num_spectral_averages)
    n_samples = filemetadata("n_samples")
    n_samples["data"] = samples * np.ones(ncvars["time"].size, dtype=np.int32)

    # 4.6 radar_parameters sub-convention -> instrument_parameters dict
    # this section needed multiple changes and/or additions since the
    # radar instrument parameters were primarily located in the global
    # attributes
    # this section is likely still incomplete
    instrument_parameters = {
        "frequency": frequency,
        "prt_mode": prt_mode,
        "prt": prt,
        "nyquist_velocity": nyquist_velocity,
        "n_samples": n_samples,
    }

    # 4.7 lidar_parameters sub-convention -> skip
    # 4.8 radar_calibration sub-convention -> skip

    # close NetCDF object
    ncobj.close()

    return Radar(
        time,
        _range,
        fields,
        metadata,
        scan_type,
        latitude,
        longitude,
        altitude,
        sweep_number,
        sweep_mode,
        fixed_angle,
        sweep_start_ray_index,
        sweep_end_ray_index,
        azimuth,
        elevation,
        instrument_parameters=instrument_parameters,
    )
